"""Autogenerated module for the SHFSG QCoDeS driver."""
from typing import Any, Dict, List, Tuple, Union
from zhinst.toolkit import CommandTable, Waveforms, Sequence
from zhinst.qcodes.driver.devices.base import ZIBaseInstrument
from zhinst.qcodes.qcodes_adaptions import ZINode, ZIChannelList


class CommandTableNode(ZINode):
    """CommandTable node.

    This class implements the basic functionality of the command table allowing
    the user to load and upload their own command table.

    A dedicated class called ``CommandTable`` exists that is the preferred way
    to create a valid command table. For more information about the
    ``CommandTable`` refer to the corresponding example or the documentation
    of that class directly.

    Args:
        root: Node used for the upload of the command table
        tree: Tree (node path as tuple) of the current node
        device_type: Device type.
    """

    def __init__(self, parent, tk_object, snapshot_cache=None, zi_node=None):
        ZINode.__init__(
            self, parent, "commandtable", snapshot_cache=snapshot_cache, zi_node=zi_node
        )
        self._tk_object = tk_object

    def check_status(self) -> bool:
        """Check status of the command table.

        Returns:
            Flag if a valid command table is loaded into the device.

        Raises:
            RuntimeError: If the command table upload into the device failed.
        """
        return self._tk_object.check_status()

    def load_validation_schema(self) -> Dict[str, Any]:
        """Load device command table validation schema.

        Returns:
            JSON validation schema for the device command tables.
        """
        return self._tk_object.load_validation_schema()

    def upload_to_device(
        self,
        ct: Union[CommandTable, str, dict],
        *,
        validate: bool = False,
        check_upload: bool = True,
    ) -> None:
        """Upload command table into the device.

        The command table can either be specified through the dedicated
        ``CommandTable`` class or in a raw format, meaning a json string or json
        dict. In the case of a json string or dict the command table is
        validated by default against the schema provided by the device.

        Args:
            ct: Command table.
            validate: Flag if the command table should be validated. (Only
                applies if the command table is passed as a raw json string or
                json dict)
            check_upload: Flag if the upload should be validated by calling
                `check_status`. This is not mandatory bat strongly recommended
                since the device does not raise an error when it rejects the
                command table. This Flag is ignored when called from within a
                transaction.

        Raises:
            RuntimeError: If the command table upload into the device failed.
            zhinst.toolkit.exceptions.ValidationError: Incorrect schema.

        .. versionchanged:: 0.4.2

            New Flag `check_upload` that makes the upload check optional.
            `check_status` is only called when not in a ongoing transaction.
        """
        return self._tk_object.upload_to_device(
            ct=ct, validate=validate, check_upload=check_upload
        )

    def load_from_device(self) -> CommandTable:
        """Load command table from the device.

        Returns:
            command table.
        """
        return self._tk_object.load_from_device()


class AWGCore(ZINode):
    """AWG Core Node."""

    def __init__(self, parent, tk_object, snapshot_cache=None, zi_node=None):
        ZINode.__init__(
            self, parent, "awg", snapshot_cache=snapshot_cache, zi_node=zi_node
        )
        self._tk_object = tk_object
        if self._tk_object.commandtable:

            self.add_submodule(
                "commandtable",
                CommandTableNode(
                    self,
                    self._tk_object.commandtable,
                    zi_node=self._tk_object.commandtable.node_info.path,
                    snapshot_cache=self._snapshot_cache,
                ),
            )

    def enable_sequencer(self, *, single: bool) -> None:
        """Starts the sequencer of a specific channel.

        Warning:
            This function is synchronous and blocks until the sequencer is enabled.
            When working with multiple instruments this function is the wrong
            approach and the sequencer should be enabled asynchronously.
            (For more information please take a look at the awg example in the
            toolkit documentation.)

        Args:
            single: Flag if the sequencer should be disabled after finishing
            execution.

        Raises:
            RuntimeError: If the sequencer could not be enabled.

        .. versionchanged:: 0.5.0

            Check the acknowledged value instead of using `wait_for_state_change`.
        """
        return self._tk_object.enable_sequencer(single=single)

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.005) -> None:
        """Wait until the AWG is finished.

        Args:
            timeout: The maximum waiting time in seconds for the generator
                (default: 10).
            sleep_time: Time in seconds to wait between requesting generator
                state

        Raises:
            RuntimeError: If continuous mode is enabled
            TimeoutError: If the sequencer program did not finish within
                the specified timeout time
        """
        return self._tk_object.wait_done(timeout=timeout, sleep_time=sleep_time)

    def compile_sequencer_program(
        self, sequencer_program: Union[str, Sequence], **kwargs: Union[str, int]
    ) -> Tuple[bytes, Dict[str, Any]]:
        """Compiles a sequencer program for the specific device.

        Args:
            sequencer_program: The sequencer program to compile.

        Keyword Args:
            samplerate: Target sample rate of the sequencer. Only allowed/
                necessary for HDAWG devices. Must correspond to the samplerate
                used by the device (device.system.clocks.sampleclock.freq()).
                If not specified the function will get the value itself from
                the device. It is recommended passing the samplerate if more
                than one sequencer code is uploaded in a row to speed up the
                execution time.
            wavepath: path to directory with waveforms. Defaults to path used
                by LabOne UI or AWG Module.
            waveforms: waveform CSV files separated by ';'
            output: name of embedded ELF filename.

        Returns:
            elf: Binary ELF data for sequencer.
            extra: Extra dictionary with compiler output.

        Examples:
            >>> elf, compile_info = device.awgs[0].compile_sequencer_program(seqc)
            >>> device.awgs[0].elf.data(elf)
            >>> device.awgs[0].ready.wait_for_state_change(1)
            >>> device.awgs[0].enable(True)

        Raises:
            RuntimeError: `sequencer_program` is empty.
            RuntimeError: If the compilation failed.

        .. versionadded:: 0.4.0
        """
        return self._tk_object.compile_sequencer_program(
            sequencer_program=sequencer_program, **kwargs
        )

    def load_sequencer_program(
        self, sequencer_program: Union[str, Sequence], **kwargs: Union[str, int]
    ) -> Dict[str, Any]:
        """Compiles the given sequencer program on the AWG Core.

        Warning:
            After uploading the sequencer program one needs to wait before for
            the awg core to become ready before it can be enabled.
            The awg core indicates the ready state through its `ready` node.
            (device.awgs[0].ready() == True)

        Args:
            sequencer_program: Sequencer program to be uploaded.

        Keyword Args:
            samplerate: Target sample rate of the sequencer. Only allowed/
                necessary for HDAWG devices. Must correspond to the samplerate
                used by the device (device.system.clocks.sampleclock.freq()).
                If not specified the function will get the value itself from
                the device. It is recommended passing the samplerate if more
                than one sequencer code is uploaded in a row to speed up the
                execution time.
            wavepath: path to directory with waveforms. Defaults to path used
                by LabOne UI or AWG Module.
            waveforms: waveform CSV files separated by ';'
            output: name of embedded ELF filename.

        Examples:
            >>> compile_info = device.awgs[0].load_sequencer_program(seqc)
            >>> device.awgs[0].ready.wait_for_state_change(1)
            >>> device.awgs[0].enable(True)

        Raises:
            RuntimeError: `sequencer_program` is empty.
            RuntimeError: If the upload or compilation failed.

        .. versionadded:: 0.3.4

            `sequencer_program` does not accept empty strings

        .. versionadded:: 0.4.0

            Use offline compiler instead of AWG module to compile the sequencer
            program. This speeds of the compilation and also enables parallel
            compilation/upload.
        """
        return self._tk_object.load_sequencer_program(
            sequencer_program=sequencer_program, **kwargs
        )

    def write_to_waveform_memory(
        self, waveforms: Waveforms, indexes: list = None
    ) -> None:
        """Writes waveforms to the waveform memory.

        The waveforms must already be assigned in the sequencer program.

        Args:
            waveforms: Waveforms that should be uploaded.
            indexes: Specify a list of indexes that should be uploaded. If
                nothing is specified all available indexes in waveforms will
                be uploaded. (default = None)

        .. versionchanged:: 0.4.2

            Removed `validate` flag and functionality. The validation check is
            now done in the `Waveforms.validate` function.
        """
        return self._tk_object.write_to_waveform_memory(
            waveforms=waveforms, indexes=indexes
        )

    def read_from_waveform_memory(self, indexes: List[int] = None) -> Waveforms:
        """Read waveforms from the waveform memory.

        Args:
            indexes: List of waveform indexes to read from the device. If not
                specified all assigned waveforms will be downloaded.

        Returns:
            Waveform object with the downloaded waveforms.
        """
        return self._tk_object.read_from_waveform_memory(indexes=indexes)

    def configure_marker_and_trigger(
        self, *, trigger_in_source: str, trigger_in_slope: str, marker_out_source: str
    ) -> None:
        """Configures the trigger inputs and marker outputs of the AWG.

        Args:
            trigger_in_source: Alias for the trigger input used by the
                sequencer. For a list of available values use:
                `available_trigger_inputs`
            trigger_in_slope: Alias for the slope of the input trigger
                used by sequencer. For a list of available values use
                `available_trigger_inputs`
            marker_out_source: Alias for the marker output source used by
                the sequencer. For a list of available values use
                `available_trigger_slopes`
        """
        return self._tk_object.configure_marker_and_trigger(
            trigger_in_source=trigger_in_source,
            trigger_in_slope=trigger_in_slope,
            marker_out_source=marker_out_source,
        )

    @property
    def available_trigger_inputs(self) -> List:
        """List the available trigger sources for the sequencer."""
        return self._tk_object.available_trigger_inputs

    @property
    def available_trigger_slopes(self) -> List:
        """List the available trigger slopes for the sequencer."""
        return self._tk_object.available_trigger_slopes

    @property
    def available_marker_outputs(self) -> List:
        """List the available trigger marker outputs for the sequencer."""
        return self._tk_object.available_marker_outputs


class SGChannel(ZINode):
    """Signal Generator Channel for the SHFSG.

    :class:`SGChannel` implements basic functionality to configure SGChannel
    settings of the :class:`SHFSG` instrument.

    Args:
        device: SHFQA device object.
        session: Underlying session.
        tree: Node tree (node path as tuple) of the corresponding node.
    """

    def __init__(self, parent, tk_object, index, snapshot_cache=None, zi_node=None):
        ZINode.__init__(
            self,
            parent,
            f"sgchannel_{index}",
            snapshot_cache=snapshot_cache,
            zi_node=zi_node,
        )
        self._tk_object = tk_object
        if self._tk_object.awg:

            self.add_submodule(
                "awg",
                AWGCore(
                    self,
                    self._tk_object.awg,
                    zi_node=self._tk_object.awg.node_info.path,
                    snapshot_cache=self._snapshot_cache,
                ),
            )

    def configure_channel(
        self, *, enable: bool, output_range: int, center_frequency: float, rf_path: bool
    ) -> None:
        """Configures the RF input and output.

        Args:
            enable: Flag if the signal output should be enabled.
            output_range: Maximal range of the signal output power in dBm
            center_frequency: Center frequency before modulation
            rf_path: Flag if the RF(True) or LF(False) path should be
                configured.
        """
        return self._tk_object.configure_channel(
            enable=enable,
            output_range=output_range,
            center_frequency=center_frequency,
            rf_path=rf_path,
        )

    def configure_pulse_modulation(
        self,
        *,
        enable: bool,
        osc_index: int = 0,
        osc_frequency: float = 100000000.0,
        phase: float = 0.0,
        global_amp: float = 0.5,
        gains: tuple = (1.0, -1.0, 1.0, 1.0),
        sine_generator_index: int = 0,
    ) -> None:
        """Configure the pulse modulation.

        Configures the sine generator to digitally modulate the AWG output, for
        generating single sideband AWG signals

        Args:
            enable: Flag if the modulation should be enabled.
            osc_index: Selects which oscillator to use
            osc_frequency: Oscillator frequency used to modulate the AWG
                outputs. (default = 100e6)
            phase: Sets the oscillator phase. (default = 0.0)
            global_amp: Global scale factor for the AWG outputs. (default = 0.5)
            gains: Sets the four amplitudes used for single sideband generation.
                Default values correspond to upper sideband with a positive
                oscillator frequency. (default = (1.0, -1.0, 1.0, 1.0))
            sine_generator_index: Selects which sine generator to use on a
                given channel.
        """
        return self._tk_object.configure_pulse_modulation(
            enable=enable,
            osc_index=osc_index,
            osc_frequency=osc_frequency,
            phase=phase,
            global_amp=global_amp,
            gains=gains,
            sine_generator_index=sine_generator_index,
        )

    def configure_sine_generation(
        self,
        *,
        enable: bool,
        osc_index: int = 0,
        osc_frequency: float = 100000000.0,
        phase: float = 0.0,
        gains: tuple = (0.0, 1.0, 1.0, 0.0),
        sine_generator_index: int = 0,
    ) -> None:
        """Configures the sine generator output.

        Configures the sine generator output of a specified channel for generating
        continuous wave signals without the AWG.

        Args:
            enable: Flag if the sine generator output should be enabled.
            osc_index: Selects which oscillator to use
            osc_frequency: Oscillator frequency used by the sine generator
                (default = 100e6)
            phase: Sets the oscillator phase. (default = 0.0)
            gains: Sets the four amplitudes used for single sideband
                generation. Default values correspond to upper sideband with a
                positive oscillator frequency.
                Gains are set in the following order I/sin, I/cos, Q/sin, Q/cos.
                (default = (0.0, 1.0, 1.0, 0.0))
            sine_generator_index: Selects which sine generator to use on a given
                channel
        """
        return self._tk_object.configure_sine_generation(
            enable=enable,
            osc_index=osc_index,
            osc_frequency=osc_frequency,
            phase=phase,
            gains=gains,
            sine_generator_index=sine_generator_index,
        )

    @property
    def awg_modulation_freq(self) -> float:
        """Modulation frequency of the AWG.

        Depends on the selected oscillator.
        """
        return self._tk_object.awg_modulation_freq


class SHFSG(ZIBaseInstrument):
    """QCoDeS driver for the Zurich Instruments SHFSG."""

    def _init_additional_nodes(self):
        """Init class specific modules and parameters."""
        if self._tk_object.sgchannels:

            channel_list = ZIChannelList(
                self,
                "sgchannels",
                SGChannel,
                zi_node=self._tk_object.sgchannels.node_info.path,
                snapshot_cache=self._snapshot_cache,
            )
            for i, x in enumerate(self._tk_object.sgchannels):
                channel_list.append(
                    SGChannel(
                        self,
                        x,
                        i,
                        zi_node=self._tk_object.sgchannels[i].node_info.path,
                        snapshot_cache=self._snapshot_cache,
                    )
                )
            # channel_list.lock()
            self.add_submodule("sgchannels", channel_list)

    def factory_reset(self, *, deep: bool = True) -> None:
        """Load the factory default settings.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after loading the factory preset (default: True).
        """
        return self._tk_object.factory_reset(deep=deep)



"""Autogenerated module for the BaseModule QCoDeS driver."""
import typing as t
from zhinst.toolkit.driver.modules import ModuleType as TKModuleType
from zhinst.toolkit.driver.modules.base_module import BaseModule as TKBaseModule
from zhinst.toolkit.driver.modules.base_module import ZIModule
from zhinst.toolkit.nodetree import Node as TKNode

from zhinst.qcodes.qcodes_adaptions import (
    ZIParameter,
    NodeDict,
    ZIInstrument,
    init_nodetree,
    tk_node_to_parameter,
)

if t.TYPE_CHECKING:
    from zhinst.qcodes.driver.devices import DeviceType
    from zhinst.qcodes.session import Session


class ZIBaseModule(ZIInstrument):
    """Generic toolkit driver for a LabOne Modules.

    All module specific class are derived from this class.
    It exposes the nodetree and also implements common functions valid for all
    modules.
    It also can be used directly, e.g. for modules that have no special class
    in toolkit.


    Args:
        tk_object: Underlying zhinst-toolkit object.
        session: Session to the Data Server.
        name: Name of the module in QCoDeS.
    """

    def __init__(self, tk_object: TKModuleType, session: "Session", name: str):
        self._tk_object = tk_object
        self._session = session
        super().__init__(
            f"zi_{name}_{len(self.instances())}", tk_object.root, is_module=True
        )
        init_nodetree(self, self._tk_object, self._snapshot_cache)

        self._tk_object.root.update_nodes(
            {
                "/device": {
                    "GetParser": lambda value: self._get_device(value),
                }
            },
            raise_for_invalid_node=False,
        )

    def _get_device(self, serial: str) -> t.Union["DeviceType", str]:
        """Convert a device serial into a QCoDeS device object.

        Args:
            serial: Serial of the device

        Returns:
            QCoDeS device object. If the serial does not
                match to a connected device the serial is returned instead.
        """
        try:
            return self._session.devices[serial]
        except (RuntimeError, KeyError):
            return serial

    def _get_node(self, node: str) -> t.Union[ZIParameter, str]:
        """Convert a raw node string into a qcodes node.

        Args:
            node (str): raw node string

        Returns:
            Node: qcodes node. (if the node can not be converted the raw node
                string is returned)
        """
        tk_node = self._tk_object._get_node(node)
        if isinstance(tk_node, str):
            return tk_node
        device = self._session.devices[tk_node.root.prefix_hide]
        return tk_node_to_parameter(device, tk_node)

    @staticmethod
    def _set_node(signal: t.Union[ZIParameter, TKNode, str]) -> str:
        """Convert a toolkit node into a raw node string.

        Args:
            signal (Union[Node,str]): node

        Returns:
            str: raw string node
        """
        try:
            node = signal.zi_node  # type: ignore[union-attr]
        except AttributeError:
            node = TKBaseModule._set_node(signal)
        return node

    def subscribe(self, signal: t.Union[ZIParameter, str]):
        """Subscribe to a node.

        The node can either be a node of this module or of a connected device.

        Args:
            signal (Node): node that should be subscribed.
        """
        try:
            self._tk_object.subscribe(signal.zi_node)  # type: ignore[union-attr]
        except AttributeError:
            self._tk_object.subscribe(signal)

    def unsubscribe(self, signal: t.Union[ZIParameter, str]):
        """Unsubscribe from a node.

        The node can either be a node of this module or of a connected device.

        Args:
            signal (Node): node that should be unsubscribe.
        """
        try:
            self._tk_object.unsubscribe(signal.zi_node)  # type: ignore[union-attr]
        except AttributeError:
            self._tk_object.unsubscribe(signal)

    @property
    def raw_module(self) -> ZIModule:  # type: ignore [type-var]
        """Underlying zhinst.core module."""
        return self._tk_object.raw_module

    def wait_done(self, *, timeout: float = 20.0, sleep_time: float = 0.5) -> None:
        """Waits until the module is finished.

        Warning: Only usable for modules that make use of the `/finished` node.

        Args:
            timeout (float): The maximum waiting time in seconds for the
                measurement (default: 20).
            sleep_time (int): Time in seconds to wait between
                requesting sweeper state. (default: 0.5)

        Raises:
            TimeoutError: The measurement is not completed before
                timeout.
        """
        return self._tk_object.wait_done(timeout=timeout, sleep_time=sleep_time)

    def progress(self) -> float:
        """Progress of the execution.

        Returns:
            Progress of the execution with a number between 0 and 1
        """
        return self._tk_object.progress()

    def execute(self) -> None:
        """Start the module execution.

        Subscription or unsubscription is not possible until the execution is
        finished.

        .. versionadded:: 0.4.1
        """
        return self._tk_object.execute()

    def read(self) -> NodeDict:
        """Read scope data.

        If the recording is still ongoing only a subset of data is returned.

        Returns:
            Scope data.
        """
        return NodeDict(self._tk_object.read())



"""Autogenerated module for the PQSC QCoDeS driver."""
from typing import List, Union
from zhinst.qcodes.driver.devices.base import ZIBaseInstrument


class PQSC(ZIBaseInstrument):
    """QCoDeS driver for the Zurich Instruments PQSC."""

    def _init_additional_nodes(self):
        """Init class specific modules and parameters."""

    def arm(self, *, deep=True, repetitions: int = None, holdoff: float = None) -> None:
        """Prepare PQSC for triggering the instruments.

        This method configures the execution engine of the PQSC and
        clears the register bank. Optionally, the *number of triggers*
        and *hold-off time* can be set when specified as keyword
        arguments. If they are not specified, they are not changed.

        Note that the PQSC is disabled at the end of the hold-off time
        after sending out the last trigger. Therefore, the hold-off time
        should be long enough such that the PQSC is still enabled when
        the feedback arrives. Otherwise, the feedback cannot be processed.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after stopping the PQSC and clearing the
                register bank (default: True).
            repetitions: If specified, the number of triggers sent
                over ZSync ports will be set (default: None).
            holdoff: If specified, the time between repeated
                triggers sent over ZSync ports will be set. It has a
                minimum value and a granularity of 100 ns
                (default: None).

        """
        return self._tk_object.arm(deep=deep, repetitions=repetitions, holdoff=holdoff)

    def run(self, *, deep: bool = True) -> None:
        """Start sending out triggers.

        This method activates the trigger generation to trigger all
        connected instruments over ZSync ports.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after enabling the PQSC (default: True).

        """
        return self._tk_object.run(deep=deep)

    def arm_and_run(self, *, repetitions: int = None, holdoff: float = None) -> None:
        """Arm the PQSC and start sending out triggers.

        Simply combines the methods arm and run. A synchronization
        is performed between the device and the data server after
        arming and running the PQSC.

        Args:
            repetitions: If specified, the number of triggers sent
                over ZSync ports will be set (default: None).
            holdoff: If specified, the time between repeated
                triggers sent over ZSync ports will be set. It has a
                minimum value and a granularity of 100 ns
                (default: None).

        """
        return self._tk_object.arm_and_run(repetitions=repetitions, holdoff=holdoff)

    def stop(self, *, deep: bool = True) -> None:
        """Stop the trigger generation.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after disabling the PQSC (default: True).

        """
        return self._tk_object.stop(deep=deep)

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.005) -> None:
        """Wait until trigger generation and feedback processing is done.

        Args:
            timeout: The maximum waiting time in seconds for the
                PQSC (default: 10).
            sleep_time: Time in seconds to wait between
                requesting PQSC state

        Raises:
            TimeoutError: If the PQSC is not done sending out all
                triggers and processing feedback before the timeout.
        """
        return self._tk_object.wait_done(timeout=timeout, sleep_time=sleep_time)

    def check_ref_clock(self, *, timeout: int = 30, sleep_time: int = 1) -> bool:
        """Check if reference clock is locked successfully.

        Args:
            timeout: Maximum time in seconds the program waits
                (default: 30).
            sleep_time: Time in seconds to wait between
                requesting the reference clock status (default: 1)

        Raises:
            TimeoutError: If the process of locking to the reference clock
                exceeds the specified timeout.
        """
        return self._tk_object.check_ref_clock(timeout=timeout, sleep_time=sleep_time)

    def check_zsync_connection(
        self,
        ports: Union[List[int], int] = 0,
        *,
        timeout: int = 30,
        sleep_time: int = 1,
    ) -> Union[List[bool], bool]:
        """Check if the ZSync connection on the given port(s) is established.

        This function checks the current status of the instrument connected to
        the given ports.

        Args:
            ports: The port numbers to check the ZSync connection for.
                It can either be a single port number given as integer or a list
                of several port numbers. (default: 0)
            timeout: Maximum time in seconds the program waits (default: 30).
            sleep_time: Time in seconds to wait between requesting the reference
                clock status (default: 1)

        Raises:
            TimeoutError: If the process of establishing a ZSync connection on
                one of the specified ports exceeds the specified timeout.
        """
        return self._tk_object.check_zsync_connection(
            ports=ports, timeout=timeout, sleep_time=sleep_time
        )

    def find_zsync_worker_port(self, device: ZIBaseInstrument) -> int:
        """Find the ID of the PQSC ZSync port connected to a given device.

        Args:
            pqsc: PQSC device over whose ports the research shall be done.
            device: device for which the connected ZSync port shall be found.

        Returns:
            Integer value represent the ID of the searched PQSC Zsync port.

        Raises:
            ToolkitError: If the given device doesn't appear to be connected
                to the PQSC via ZSync.

        .. versionadded:: 0.5.1
        """
        return self._tk_object.find_zsync_worker_port(device=device._tk_object)



import pytest
from unittest.mock import patch
from pathlib import Path
from zhinst.qcodes import ZISession


@pytest.fixture()
def data_dir(request):
    yield Path(request.fspath).parent / "data"


@pytest.fixture()
def mock_connection():
    with patch("zhinst.toolkit.session.core.ziDAQServer", autospec=True) as connection:
        yield connection


@pytest.fixture()
def session(data_dir, mock_connection):
    json_path = data_dir / "nodedoc_zi.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json
    yield ZISession("localhost")



"""A script to generate Notebooks for documentation."""
import argparse
import fnmatch
import os
import typing as t
from pathlib import Path

import requests
from jupytext import cli as jupytext_cli

BASE_EXAMPLE_URL = "https://docs.zhinst.com/zhinst-qcodes/en/latest/examples"
EXAMPLES_DIR = Path(__file__).parent.parent / "examples"
EXAMPLES_ONLY_SYNC = ["nodetree.md"]
EXCLUDED_FILES = ["README.md"]


def download_example_file(filename: str) -> bytes:
    """Download example file.

    Args:
        filename: notebook filename.

    Returns:
        Notebook contents
    """
    url = f"{BASE_EXAMPLE_URL}/{filename}"
    response = requests.get(url)
    response.raise_for_status()
    return response.content


def get_notebook_examples() -> None:
    """Get notebook examples."""
    for example_file in fnmatch.filter(os.listdir(EXAMPLES_DIR), "*.md"):
        if example_file in EXAMPLES_ONLY_SYNC or example_file in EXCLUDED_FILES:
            continue
        example_file = example_file.replace(".md", ".ipynb")
        contents = download_example_file(example_file)
        if contents:
            with open(f"{EXAMPLES_DIR / example_file}", "wb") as f:
                f.write(contents)


def generate_and_sync_example_notebooks(src: t.List[Path]) -> None:
    """Generate and sync given source files to notebooks.

    Args:
        src: Source files
    """
    str_path = [str(path) for path in src]
    jupytext_cli.jupytext(["--sync", *str_path])


def generate_notebooks(args: argparse.Namespace) -> None:
    """Generate notebooks either from local or remote.

    Args:
        args: Namespace arguments
    """
    if args.src == "local":
        generate_and_sync_example_notebooks([EXAMPLES_DIR / "*.md"])
    else:
        get_notebook_examples()
        generate_and_sync_example_notebooks(
            [EXAMPLES_DIR / file for file in EXAMPLES_ONLY_SYNC]
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate Notebooks.")
    parser.add_argument("src", help="Source of Notebooks", choices=["local", "remote"])
    generate_notebooks(parser.parse_args())



"""QCoDeS Drivers for Zurich Instruments devices."""

from zhinst.qcodes.session import ZISession
from zhinst.qcodes.device_creator import (
    HDAWG,
    MFLI,
    MFIA,
    PQSC,
    SHFQA,
    SHFQC,
    SHFSG,
    UHFLI,
    UHFQA,
    ZIDevice,
    HF2,
)

from zhinst.toolkit import (
    Waveforms,
    CommandTable,
    Sequence,
    PollFlags,
    AveragingMode,
    SHFQAChannelMode,
)

try:
    from zhinst.qcodes._version import version as __version__
except ModuleNotFoundError:
    pass

__all__ = [
    "ZISession",
    "HDAWG",
    "MFLI",
    "MFIA",
    "PQSC",
    "SHFQA",
    "SHFQC",
    "SHFSG",
    "UHFLI",
    "UHFQA",
    "ZIDevice",
    "HF2",
    "Waveforms",
    "CommandTable",
    "Sequence",
    "PollFlags",
    "AveragingMode",
    "SHFQAChannelMode",
]



import pytest
from unittest.mock import patch
from fixtures import mock_connection, data_dir, session


class TestDataServerSession:
    def test_setup(self, mock_connection, session):
        mock_connection.assert_called_once_with("localhost", 8004, 6)
        mock_connection.return_value.listNodesJSON.assert_called_once_with("/zi/*")
        assert not session.is_hf2_server
        assert session.server_host == mock_connection.return_value.host
        assert session.server_port == mock_connection.return_value.port



"""Autogenerated module for the ScopeModule QCoDeS driver."""
import typing as t
from zhinst.toolkit.driver.modules.scope_module import ScopeModule as TKScopeModule

from zhinst.qcodes.driver.modules.base_module import ZIBaseModule


if t.TYPE_CHECKING:
    from zhinst.qcodes.session import Session


class ZIScopeModule(ZIBaseModule):
    """Scope Module.

    The Scope Module corresponds to the functionality available in the Scope
    tab in the LabOne User Interface and provides API users with an interface
    to acquire assembled and scaled scope data from the instrument
    programmatically.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/scope_module.html


    Although it is possible to acquire scope data using the lower-level
    subscribe/poll method, the Scope Module provides API users with several
    advantages. Specifically, the Scope Module:

    * Provides a uniform interface to acquire scope data from all instrument
      classes (HF2 scope usage differs from and MF and UHF devices, especially
      with regards to scaling).
    * Scales and offsets the scope wave data to get physically meaningful
      values. If data is polled from the device node using subscribe/poll the
      scaling and offset must be applied manually.
    * Assembles large multi-block transferred scope data into single complete
      records. When the scope is configured to record large scope lengths and
      data is directly polled from the device node /DEV…​/SCOPES/n/WAVE the data
      is split into multiple blocks for efficient transfer of data from the
      Data Server to the API; these must then be programmatically reassembled.
      The Scope Module performs this assembly and returns complete scope
      records (unless used in pass-through mode, mode=0).
    * Can be configured to return the FFT of the acquired scope records
      (with mode=3) as provided by the Scope Tab in the LabOne UI. FFT data is
      not available from the device nodes in the /DEV/…​./SCOPES/ branch using
      subscribe/poll.
    * Can be configured to average the acquired scope records the
      averager/parameters.
    * Can be configured to return a specific number of scope records using the
      historylength parameter.


    Args:
        tk_object: Underlying zhinst-toolkit object.
        session: Session to the Data Server.
        name: Name of the module in QCoDeS.
    """

    def __init__(
        self, tk_object: TKScopeModule, session: "Session", name: str = "scope_module"
    ):
        super().__init__(tk_object, session, name)

    def finish(self) -> None:
        """Stop the module."""
        return self._tk_object.finish()



"""Module for toolkit representations of native LabOne modules."""
import typing as t

from zhinst.qcodes.driver.modules.base_module import ZIBaseModule
from zhinst.qcodes.driver.modules.daq_module import ZIDAQModule
from zhinst.qcodes.driver.modules.device_settings_module import ZIDeviceSettingsModule
from zhinst.qcodes.driver.modules.impedance_module import ZIImpedanceModule
from zhinst.qcodes.driver.modules.pid_advisor_module import ZIPIDAdvisorModule
from zhinst.qcodes.driver.modules.precompensation_advisor_module import (
    ZIPrecompensationAdvisorModule,
)
from zhinst.qcodes.driver.modules.scope_module import ZIScopeModule
from zhinst.qcodes.driver.modules.shfqa_sweeper import ZISHFQASweeper
from zhinst.qcodes.driver.modules.sweeper_module import ZISweeperModule

ModuleType = t.Union[
    ZIBaseModule,
    ZIDAQModule,
    ZIDeviceSettingsModule,
    ZIImpedanceModule,
    ZIPIDAdvisorModule,
    ZIPrecompensationAdvisorModule,
    ZIScopeModule,
    ZISHFQASweeper,
    ZISweeperModule,
]

__all__ = [
    "ModuleType",
    "ZIBaseModule",
    "ZIDAQModule",
    "ZIDeviceSettingsModule",
    "ZIImpedanceModule",
    "ZIPIDAdvisorModule",
    "ZIPrecompensationAdvisorModule",
    "ZIScopeModule",
    "ZISHFQASweeper",
    "ZISweeperModule",
]



"""Configuration for zhinst QCoDes driver generation."""
from pathlib import Path


PKG_ROOT = Path(__file__).parent
TEMPLATE_PATH = PKG_ROOT / "templates"

# Device drivers
OUTPUT_DIR_DEVICES_DRIVER = PKG_ROOT.parent / "src/zhinst/qcodes/driver/devices/"

# devices API
DEVICE_DRIVERS = ["SHFQA", "SHFSG", "SHFQC", "HDAWG", "PQSC", "UHFLI", "UHFQA"]
TOOLKIT_DEVICE_MODULE = "zhinst.toolkit.driver.devices"

# Module drivers
OUTPUT_DIR_MODULE_DRIVER = PKG_ROOT.parent / "src/zhinst/qcodes/driver/modules/"

# module API
MODULE_DRIVERS = [
    "BaseModule",
    "DAQModule",
    "ScopeModule",
    "SweeperModule",
    "ImpedanceModule",
    "DeviceSettingsModule",
    "PIDAdvisorModule",
    "PrecompensationAdvisorModule",
]
TOOLKIT_MODULE_MODULE = "zhinst.toolkit.driver.modules"

# Typing
# Weird typing infos that can be replaced with the right term
TYPE_HINT_REPLACEMENTS = {
    "<built-in function array>": "np.array",
    "~Numpy2DArray": "np.ndarray",
}



"""Autogenerated module for the DeviceSettingsModule QCoDeS driver."""
import typing as t
from typing import Union
from pathlib import Path
from zhinst.toolkit.driver.modules.device_settings_module import (
    DeviceSettingsModule as TKDeviceSettingsModule,
)

from zhinst.qcodes.driver.modules.base_module import ZIBaseModule

from zhinst.qcodes.qcodes_adaptions import (
    NodeDict,
)

if t.TYPE_CHECKING:
    from zhinst.qcodes.driver.devices import DeviceType
    from zhinst.qcodes.session import Session


class ZIDeviceSettingsModule(ZIBaseModule):
    """Implements the device settings module for storing and loading settings.

    The Device Settings Module provides functionality for saving and loading
    device settings to and from file. The file is saved in XML format.

    For simple save and load two helper functions exist `save_to_file` and
    `load_from_file`.

    Note: It is not recommend to use this function to read the
        device settings. Instead one can use the zhinst-toolkit functionality
        to read all settings from a device/subtree from the device directly by
        calling it.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/device_settings_module.html


    Args:
        tk_object: Underlying zhinst-toolkit object.
        session: Session to the Data Server.
        name: Name of the module in QCoDeS.
    """

    def __init__(
        self,
        tk_object: TKDeviceSettingsModule,
        session: "Session",
        name: str = "device_settings_module",
    ):
        super().__init__(tk_object, session, name)

    def load_from_file(
        self,
        filename: Union[str, Path],
        device: Union["DeviceType", str],
        timeout: float = 30,
    ) -> None:
        """Load a LabOne settings file to a device.

        This function creates an new module instance to avoid misconfiguration.
        It is also synchronous, meaning it will block until loading the
        settings has finished.

        Args:
            filename: The path to the settings file.
            device: The device to load the settings to.
            timeout: Max time to wait for the loading to finish.

        Raises:
            TimeoutError: If the loading of the settings timed out.
        """
        return self._tk_object.load_from_file(
            filename=filename, device=device, timeout=timeout
        )

    def save_to_file(
        self,
        filename: Union[str, Path],
        device: Union["DeviceType", str],
        timeout: int = 30,
    ) -> None:
        """Save the device settings to a LabOne settings file.

        This function creates an new module instance to avoid misconfiguration.
        It is also synchronous, meaning it will block until save operation has
        finished.

        Args:
            filename: The path to the settings file.
            device: The device which settings should be saved.
            timeout: Max time to wait for the loading to finish.

        Raises:
            TimeoutError: If the loading of the settings timed out.
        """
        return self._tk_object.save_to_file(
            filename=filename, device=device, timeout=timeout
        )

    def read(self) -> NodeDict:
        """Read device settings.

        Note: It is not recommend to use this function to read the
        device settings. Instead one can use the zhinst-toolkit functionality
        to read all settings from a device/subtree from the device directly by
        calling it.

        >>> device = session.connect_device()
        >>> ...
        >>> device()
        <all device settings>
        >>> device.demods()
        <all demodulator settings>

        Returns:
            Device settings.
        """
        return NodeDict(self._tk_object.read())



"""Connection Manager for the LabOne Python API."""

from collections.abc import MutableMapping
from functools import cached_property
import typing as t

from zhinst.toolkit.session import Devices as TKDevices
from zhinst.toolkit.session import PollFlags
from zhinst.toolkit.session import Session as TKSession
from zhinst.toolkit.session import ModuleHandler as TKModuleHandler
from zhinst.core import ziDAQServer

import zhinst.qcodes.driver.devices as ZIDevices
import zhinst.qcodes.driver.modules as ZIModules
from zhinst.qcodes.qcodes_adaptions import (
    init_nodetree,
    tk_node_to_parameter,
    ZIParameter,
    ZIInstrument,
)


class Devices(MutableMapping):
    """Mapping class for the connected devices.

    Maps the connected devices from data server to lazy device objects.
    On every access the connected devices are read from the data server. This
    ensures that even if devices get connected/disconnected through another
    session the list will be up to date.

    Args:
        session: active session to the data server.
        tk_devices: toolkit devices object.
    """

    def __init__(self, session: "Session", tk_devices: TKDevices):
        self._tk_devices = tk_devices
        self._session = session
        self._devices: t.Dict[str, ZIDevices.DeviceType] = {}
        self._default_properties: t.Dict[
            str, t.Tuple[t.Optional[str], t.Optional[bool]]
        ] = {}

    def __getitem__(self, key) -> ZIDevices.DeviceType:
        key = key.lower()
        if key in self.connected():
            if key not in self._devices:
                tk_device = self._tk_devices[key]
                name, raw = self._default_properties.get(key, (None, False))
                self._devices[key] = ZIDevices.DEVICE_CLASS_BY_MODEL.get(
                    tk_device.__class__.__name__, ZIDevices.ZIBaseInstrument
                )(tk_device, self._session, name=name, raw=raw)
            return self._devices[key]
        raise KeyError(key)

    def __setitem__(self, key: str, device: ZIDevices.DeviceType) -> None:
        if device.serial not in self.connected():
            raise LookupError(
                "Illegal operation. Devices must be connected through the session."
            )
        self._devices[key] = device

    def __delitem__(self, key):
        self._devices.pop(key, None)

    def __iter__(self):
        return iter(self.connected())

    def __len__(self):
        return len(self.connected())

    def update_device_properties(
        self, serial: str, name: t.Optional[str], raw: t.Optional[bool]
    ) -> None:
        """Update the properties for a device.

        The device options are used when the QCoDeS option of a new device
        is created.

        Args:
            serial: Serial of the device (e.g. dev1234)
            name: Optional name of the QCoDeS device object
            raw: Flag if qcodes instance should only created with the nodes and
                not forwarding the toolkit functions. (default = False)

        Raises:
            RuntimeError: If the device is already created
        """
        if serial in self._devices:
            raise RuntimeError(
                f"The Qcodes Instance of {serial} already exists.\n"
                "The device properties can therfor no longer be changed"
            )
        self._default_properties[serial.lower()] = (name, raw)

    def connected(self) -> t.List[str]:
        """Get a list of devices connected to the data server.

        Returns:
            list[str]: List of all connected devices.
        """
        return self._tk_devices.connected()

    def visible(self) -> t.List[str]:
        """Get a list of devices visible to the data server.

        Returns:
            list[str]: List of all connected devices.
        """
        return self._tk_devices.visible()


class ModuleHandler:
    """Modules of LabOne.

    Handler for all additional so called modules by LabOne. A LabOne module is
    bound to a user session but creates a independent session to the Data Server.
    This has the advantage that they do not interfere with the user session. It
    also means that creating a session causes additional resources allocation,
    both at the client and the data server. New modules should therefore only be
    instantiated with care.

    Toolkit holds a lazy generated instance of all modules. This ensures that
    not more than one modules of each type gets created by accident and that the
    access to the modules is optimized.

    Of course there are many use cases where more than one module of a single
    type is required. This class therefore also exposes a ``create`` function for
    each LabOne module. These functions create a unmanaged instance of that
    module (unmanaged means toolkit does not hold an instance of that module).

    Args:
        session: Active user session
        tk_modules: Underlying toolkit module handler
    """

    def __init__(self, session: "Session", tk_modules: TKModuleHandler):
        self._session = session
        self._tk_modules = tk_modules

    def create_awg_module(self) -> ZIModules.ZIBaseModule:
        """Create a QCoDeS instance of the AWGModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `awg_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_awg_module()
        return ZIModules.ZIBaseModule(module, self._session, name="awg_module")

    def create_daq_module(self) -> ZIModules.ZIDAQModule:
        """Create a QCoDeS instance of the DAQModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `awg_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_daq_module()
        return ZIModules.ZIDAQModule(module, self._session)

    def create_device_settings_module(self) -> ZIModules.ZIDeviceSettingsModule:
        """Create a QCoDeS instance of the DeviceSettingsModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `awg_module`.

        Returns:
            DeviceSettingsModule: created module
        """
        module = self._tk_modules.create_device_settings_module()
        return ZIModules.ZIDeviceSettingsModule(module, self._session)

    def create_impedance_module(self) -> ZIModules.ZIImpedanceModule:
        """Create a QCoDeS instance of the ImpedanceModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `impedance_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_impedance_module()
        return ZIModules.ZIImpedanceModule(module, self._session)

    def create_mds_module(self) -> ZIModules.ZIBaseModule:
        """Create a QCoDeS instance of the PIDAdvisorModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `multi_device_sync_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_mds_module()
        return ZIModules.ZIBaseModule(module, self._session, name="mds_module")

    def create_pid_advisor_module(self) -> ZIModules.ZIPIDAdvisorModule:
        """Create a QCoDeS instance of the PIDAdvisorModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `pid_advisor_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_pid_advisor_module()
        return ZIModules.ZIPIDAdvisorModule(module, self._session)

    def create_precompensation_advisor_module(
        self,
    ) -> ZIModules.ZIPrecompensationAdvisorModule:
        """Create a QCoDeS instance of the PrecompensationAdvisorModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `precompensation_advisor_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_precompensation_advisor_module()
        return ZIModules.ZIPrecompensationAdvisorModule(module, self._session)

    def create_qa_module(self) -> ZIModules.ZIBaseModule:
        """Create a QCoDeS instance of the AwgModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `qa_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_qa_module()
        return ZIModules.ZIBaseModule(module, self._session, name="qa_module")

    def create_scope_module(self) -> ZIModules.ZIScopeModule:
        """Create a QCoDeS instance of the AwgModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `scope_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_scope_module()
        return ZIModules.ZIScopeModule(module, self._session)

    def create_sweeper_module(self) -> ZIModules.ZISweeperModule:
        """Create a QCoDeS instance of the SweeperModule.

        The new instance creates a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `sweeper_module`.

        Returns:
            created module
        """
        module = self._tk_modules.create_sweeper_module()
        return ZIModules.ZISweeperModule(module, self._session)

    def create_shfqa_sweeper(self) -> ZIModules.ZISHFQASweeper:
        """Create an instance of the SHFQASweeper.

        For now the general sweeper module does not support the SHFQA. However a
        python based implementation called ``SHFSweeper`` does already provide
        this functionality. The ``SHFSweeper`` is part of the ``zhinst`` module
        and can be found in the utils.

        Toolkit wraps around the ``SHFSweeper`` and exposes a interface that is
        similar to the LabOne modules, meaning the parameters are exposed in a
        node tree like structure.

        In addition a new session is created. This has the benefit that the
        sweeper implementation does not interfere with the the commands and
        setups from the user.

        Returns:
            created object
        """
        module = self._tk_modules.create_shfqa_sweeper()
        return ZIModules.ZISHFQASweeper(module, self._session)

    @cached_property
    def awg(self) -> ZIModules.ZIBaseModule:
        """Managed instance of the zhinst.core.AwgModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_awg_module()

    @cached_property
    def daq(self) -> ZIModules.ZIBaseModule:
        """Managed instance of the zhinst.core.DataAcquisitionModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_daq_module()

    @cached_property
    def device_settings(self) -> ZIModules.ZIDeviceSettingsModule:
        """Managed instance of the zhinst.core.DeviceSettingsModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_device_settings_module()

    @cached_property
    def impedance(self) -> ZIModules.ZIImpedanceModule:
        """Managed instance of the zhinst.core.ImpedanceModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_impedance_module()

    @cached_property
    def mds(self) -> ZIModules.ZIBaseModule:
        """Managed instance of the zhinst.core.MultiDeviceSyncModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_mds_module()

    @cached_property
    def pid_advisor(self) -> ZIModules.ZIPIDAdvisorModule:
        """Managed instance of the zhinst.core.PidAdvisorModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_pid_advisor_module()

    @cached_property
    def precompensation_advisor(self) -> ZIModules.ZIPrecompensationAdvisorModule:
        """Managed instance of the zhinst.core.PrecompensationAdvisorModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_precompensation_advisor_module()

    @cached_property
    def qa(self) -> ZIModules.ZIBaseModule:
        """Managed instance of the zhinst.core.QuantumAnalyzerModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_qa_module()

    @cached_property
    def scope(self) -> ZIModules.ZIScopeModule:
        """Managed instance of the zhinst.core.ScopeModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_scope_module()

    @cached_property
    def sweeper(self) -> ZIModules.ZISweeperModule:
        """Managed instance of the zhinst.core.SweeperModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_sweeper_module()

    @cached_property
    def shfqa_sweeper(self) -> ZIModules.ZISHFQASweeper:
        """Managed instance of the zhinst.core.SweeperModule.

        Managed in this sense means that only one instance is created
        and hold inside the connection Manager. This makes it easier to access
        the modules from with toolkit, since creating a module requires
        resources.
        """
        return self.create_shfqa_sweeper()


class ZISession:
    """Session to a data server.

    Zurich Instruments devices use a server-based connectivity methodology.
    Server-based means that all communication between the user and the
    instrument takes place via a computer program called a server, the data
    sever. The data sever recognizes available instruments and manages all
    communication between the instrument and the host computer on one side, and
    communication to all the connected clients on the other side. (For more
    information on the architecture please refer to the user manual
    http://docs.zhinst.com/labone_programming_manual/introduction.html)

    The entry point into any connection is therefor a client session to a
    existing data sever. This class represents a single client session to a
    data server. The session enables the user to connect to one or multiple
    instruments (also creates the dedicated objects for each device), access
    the LabOne modules and poll data.

    Since QCoDeS normally instantiate the device specific objects directly
    this driver also exposes helper classes for that directly. These helper
    classes create a session and connect the specified device to it. To avoid
    that each device has a own session by default ``ZISession`` only creates one
    session to a single data server and reuses that.

    Info:
        Except for the HF2 a single session can be used to connect to all
        devices from Zurich Instruments. Since the HF2 is historically based on
        another data server called the hf2 data server it is not possible to
        connect HF2 devices a "normal" data server and also not possible to
        connect devices apart from HF2 to the hf2 data server.

    Args:
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port 8004 (8005 for HF2 if specified).
            (default = None)
        hf2: Flag if the session should be established with an HF2 data sever or
            the "normal" one for all other devices. If not specified the session
            will detect the type of the data server based on the port.
            (default = None)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the Flag will create a new session.
            Warning: Creating a new session should be done carefully since it
            requires more resources and can create unwanted side effects.
        connection: Existing daq server object. If specified the session will
            not create a new session to the data server but reuse the passed
            one. (default = None)
    """

    def __new__(
        cls,
        server_host: str,
        server_port: t.Optional[int] = None,
        *,
        hf2: t.Optional[bool] = None,
        new_session=False,
        connection: t.Optional[ziDAQServer] = None,
    ):
        """Session creator."""
        if not new_session:
            for instance in Session.instances():
                if instance.server_host == server_host and (
                    (instance.is_hf2_server and hf2)
                    or server_port is None
                    or instance.server_port == server_port
                ):
                    return instance
        return Session(server_host, server_port, hf2=hf2, connection=connection)


class Session(ZIInstrument):
    """Session to a data server.

    Zurich Instruments devices use a server-based connectivity methodology.
    Server-based means that all communication between the user and the
    instrument takes place via a computer program called a server, the data
    sever. The data sever recognizes available instruments and manages all
    communication between the instrument and the host computer on one side, and
    communication to all the connected clients on the other side. (For more
    information on the architecture please refer to the user manual
    http://docs.zhinst.com/labone_programming_manual/introduction.html)

    The entry point into for any connection is therefor a client session to a
    existing data sever. This class represents a single client session to a
    data server. The session enables the user to connect to one or multiple
    instruments (also creates the dedicated objects for each device), access
    the LabOne modules and poll data. In short it is the only object the user
    need to create by himself.

    Info:
        Except for the HF2 a single session can be used to connect to all
        devices from Zurich Instruments. Since the HF2 is historically based on
        another data server called the hf2 data server it is not possible to
        connect HF2 devices a "normal" data server and also not possible to
        connect devices apart from HF2 to the hf2 data server.

    Args:
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port 8004 (8005 for HF2 if specified).
            (default = None)
        hf2: Flag if the session should be established with an HF2 data sever or
            the "normal" one for all other devices. If not specified the session
            will detect the type of the data server based on the port.
            (default = None)
        connection: Existing daq server object. If specified the session will
            not create a new session to the data server but reuse the passed
            one. (default = None)
    """

    def __init__(
        self,
        server_host: str,
        server_port: t.Optional[int] = None,
        *,
        hf2: t.Optional[bool] = None,
        connection: t.Optional[ziDAQServer] = None,
    ):
        self._tk_object = TKSession(
            server_host, server_port, connection=connection, hf2=hf2
        )
        super().__init__(f"zi_session_{len(self.instances())}", self._tk_object.root)
        self._devices = Devices(self, self._tk_object.devices)
        self._modules = ModuleHandler(self, self._tk_object.modules)
        init_nodetree(self, self._tk_object.root, self._snapshot_cache)

    def connect_device(
        self,
        serial: str,
        *,
        interface: t.Optional[str] = None,
        name: t.Optional[str] = None,
        raw: t.Optional[bool] = None,
    ) -> ZIDevices.DeviceType:
        """Establish a connection to a device.

        Info:
            It is allowed to call this function for an already connected device.
            In that case the function simply returns the device object of the
            already connected device.

        Args:
            serial: Serial number of the device, e.g. *'dev12000'*.
                The serial number can be found on the back panel of the
                instrument.
            interface: Device interface (e.g. = "1GbE"). If not specified
                the default interface from the discover is used.
            name: Name of the instrument in qcodes.
                (default = "zi_{dev_type}_{serial}")
            raw: Flag if qcodes instance should only created with the nodes and
                not forwarding the toolkit functions. (default = False)

        Returns:
            Device object
        """
        if name or raw is not None:
            self._devices.update_device_properties(serial, name, raw)
        self._tk_object.connect_device(serial, interface=interface)
        return self._devices[serial]

    def disconnect_device(self, serial: str) -> None:
        """Disconnect a device.

        Warning:
            This function will return immediately. The disconnection of the
            device may not yet finished.

        Args:
            serial (str): Serial number of the device, e.g. *'dev12000'*.
                The serial number can be found on the back panel of the instrument.
        """
        self._devices.pop(serial, None)
        self._tk_object.disconnect_device(serial)

    def sync(self) -> None:
        """Synchronize all connected devices.

        Synchronization in this case means creating a defined state.

        The following steps are performed:
            * Ensures that all set commands have been flushed to the device
            * Ensures that get and poll commands only return data which was
              recorded after the sync command. (ALL poll buffers are cleared!)
            * Blocks until all devices have cleared their busy flag.

        Warning:
            The sync is performed for all devices connected to the daq server

        Warning:
            This command is a blocking command that can take a substantial
            amount of time.

        Raises:
            RuntimeError: ZIAPIServerException: Timeout during sync of device
        """
        self._tk_object.sync()

    def poll(
        self,
        recording_time: float = 0.1,
        timeout: float = 0.5,
        flags: PollFlags = PollFlags.DEFAULT,
    ) -> t.Dict[ZIParameter, t.Dict[str, t.Any]]:
        """Polls all subscribed data.

        Poll the value changes in all subscribed nodes since either subscribing
        or the last poll (assuming no buffer overflow has occurred on the Data
        Server).

        Args:
            recording_time: defines the duration of the poll. (Note that not
                only the newly recorder values are polled but all values since
                either subscribing or the last pill). Needs to be larger than
                zero. (default = 0.1)
            timeout: Adds an additional timeout in seconds on top of
                `recording_time`. Only relevant when communicating in a slow
                network. In this case it may be set to a value larger than the
                expected round-trip time in the network. (default = 0.5)
            flags: Flags for the polling (see :class `PollFlags`:)

        Returns:
            Polled data in a dictionary. The key is a `Node` object and the
            value is a dictionary with the raw data from the device
        """
        polled_data_tk = self._tk_object.poll(
            recording_time=recording_time, timeout=timeout, flags=flags
        )
        polled_data = {}
        for tk_node, data in polled_data_tk.items():
            tk_node = self._tk_object.raw_path_to_node(tk_node)
            device = self.devices[tk_node.root.prefix_hide]
            parameter = tk_node_to_parameter(device, tk_node)
            polled_data[parameter] = data
        return polled_data

    @property
    def devices(self) -> Devices:
        """Mapping for the connected devices."""
        return self._devices

    @property
    def modules(self) -> ModuleHandler:
        """Modules of LabOne."""
        return self._modules

    @property
    def is_hf2_server(self) -> bool:
        """Flag if the data server is a HF2 Data Server."""
        return self._tk_object.is_hf2_server

    @property
    def daq_server(self) -> ziDAQServer:
        """Managed instance of the zi.ziDAQServer."""
        return self._tk_object.daq_server

    @property
    def server_host(self) -> str:
        """Server host."""
        return self._tk_object.server_host

    @property
    def server_port(self) -> int:
        """Server port."""
        return self._tk_object.server_port

    @property
    def toolkit_session(self) -> TKSession:
        """Underlying zhinst-toolkit session."""
        return self._tk_object



def test_api_all():
    from zhinst.qcodes import (
        ZIDevice,
        ZISession,
        SHFQA,
        SHFQC,
        SHFSG,
        UHFLI,
        UHFQA,
        HDAWG,
        HF2,
        PQSC,
        MFIA,
        MFLI,
    )



"""Autogenerated module for the UHFQA QCoDeS driver."""
from typing import Union, Optional, List, Dict, Any, Tuple
import numpy as np
from zhinst.toolkit import CommandTable, Waveforms, Sequence
from zhinst.qcodes.driver.devices.base import ZIBaseInstrument
from zhinst.qcodes.qcodes_adaptions import ZINode, ZIChannelList


class CommandTableNode(ZINode):
    """CommandTable node.

    This class implements the basic functionality of the command table allowing
    the user to load and upload their own command table.

    A dedicated class called ``CommandTable`` exists that is the preferred way
    to create a valid command table. For more information about the
    ``CommandTable`` refer to the corresponding example or the documentation
    of that class directly.

    Args:
        root: Node used for the upload of the command table
        tree: Tree (node path as tuple) of the current node
        device_type: Device type.
    """

    def __init__(self, parent, tk_object, snapshot_cache=None, zi_node=None):
        ZINode.__init__(
            self, parent, "commandtable", snapshot_cache=snapshot_cache, zi_node=zi_node
        )
        self._tk_object = tk_object

    def check_status(self) -> bool:
        """Check status of the command table.

        Returns:
            Flag if a valid command table is loaded into the device.

        Raises:
            RuntimeError: If the command table upload into the device failed.
        """
        return self._tk_object.check_status()

    def load_validation_schema(self) -> Dict[str, Any]:
        """Load device command table validation schema.

        Returns:
            JSON validation schema for the device command tables.
        """
        return self._tk_object.load_validation_schema()

    def upload_to_device(
        self,
        ct: Union[CommandTable, str, dict],
        *,
        validate: bool = False,
        check_upload: bool = True,
    ) -> None:
        """Upload command table into the device.

        The command table can either be specified through the dedicated
        ``CommandTable`` class or in a raw format, meaning a json string or json
        dict. In the case of a json string or dict the command table is
        validated by default against the schema provided by the device.

        Args:
            ct: Command table.
            validate: Flag if the command table should be validated. (Only
                applies if the command table is passed as a raw json string or
                json dict)
            check_upload: Flag if the upload should be validated by calling
                `check_status`. This is not mandatory bat strongly recommended
                since the device does not raise an error when it rejects the
                command table. This Flag is ignored when called from within a
                transaction.

        Raises:
            RuntimeError: If the command table upload into the device failed.
            zhinst.toolkit.exceptions.ValidationError: Incorrect schema.

        .. versionchanged:: 0.4.2

            New Flag `check_upload` that makes the upload check optional.
            `check_status` is only called when not in a ongoing transaction.
        """
        return self._tk_object.upload_to_device(
            ct=ct, validate=validate, check_upload=check_upload
        )

    def load_from_device(self) -> CommandTable:
        """Load command table from the device.

        Returns:
            command table.
        """
        return self._tk_object.load_from_device()


class AWG(ZINode):
    """AWG node.

    This class implements the basic functionality for the device specific
    arbitrary waveform generator.
    Besides the upload/compilation of sequences it offers the upload of
    waveforms and command tables.

    Args:
        root: Root of the nodetree
        tree: Tree (node path as tuple) of the current node
        session: Underlying session.
        serial: Serial of the device.
        index: Index of the corresponding awg channel
        device_type: Device type
    """

    def __init__(self, parent, tk_object, index, snapshot_cache=None, zi_node=None):
        ZINode.__init__(
            self, parent, f"awg_{index}", snapshot_cache=snapshot_cache, zi_node=zi_node
        )
        self._tk_object = tk_object
        if self._tk_object.commandtable:

            self.add_submodule(
                "commandtable",
                CommandTableNode(
                    self,
                    self._tk_object.commandtable,
                    zi_node=self._tk_object.commandtable.node_info.path,
                    snapshot_cache=self._snapshot_cache,
                ),
            )

    def enable_sequencer(self, *, single: bool) -> None:
        """Starts the sequencer of a specific channel.

        Warning:
            This function is synchronous and blocks until the sequencer is enabled.
            When working with multiple instruments this function is the wrong
            approach and the sequencer should be enabled asynchronously.
            (For more information please take a look at the awg example in the
            toolkit documentation.)

        Args:
            single: Flag if the sequencer should be disabled after finishing
            execution.

        Raises:
            RuntimeError: If the sequencer could not be enabled.

        .. versionchanged:: 0.5.0

            Check the acknowledged value instead of using `wait_for_state_change`.
        """
        return self._tk_object.enable_sequencer(single=single)

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.005) -> None:
        """Wait until the AWG is finished.

        Args:
            timeout: The maximum waiting time in seconds for the generator
                (default: 10).
            sleep_time: Time in seconds to wait between requesting generator
                state

        Raises:
            RuntimeError: If continuous mode is enabled
            TimeoutError: If the sequencer program did not finish within
                the specified timeout time
        """
        return self._tk_object.wait_done(timeout=timeout, sleep_time=sleep_time)

    def compile_sequencer_program(
        self, sequencer_program: Union[str, Sequence], **kwargs: Union[str, int]
    ) -> Tuple[bytes, Dict[str, Any]]:
        """Compiles a sequencer program for the specific device.

        Args:
            sequencer_program: The sequencer program to compile.

        Keyword Args:
            samplerate: Target sample rate of the sequencer. Only allowed/
                necessary for HDAWG devices. Must correspond to the samplerate
                used by the device (device.system.clocks.sampleclock.freq()).
                If not specified the function will get the value itself from
                the device. It is recommended passing the samplerate if more
                than one sequencer code is uploaded in a row to speed up the
                execution time.
            wavepath: path to directory with waveforms. Defaults to path used
                by LabOne UI or AWG Module.
            waveforms: waveform CSV files separated by ';'
            output: name of embedded ELF filename.

        Returns:
            elf: Binary ELF data for sequencer.
            extra: Extra dictionary with compiler output.

        Examples:
            >>> elf, compile_info = device.awgs[0].compile_sequencer_program(seqc)
            >>> device.awgs[0].elf.data(elf)
            >>> device.awgs[0].ready.wait_for_state_change(1)
            >>> device.awgs[0].enable(True)

        Raises:
            RuntimeError: `sequencer_program` is empty.
            RuntimeError: If the compilation failed.

        .. versionadded:: 0.4.0
        """
        return self._tk_object.compile_sequencer_program(
            sequencer_program=sequencer_program, **kwargs
        )

    def load_sequencer_program(
        self, sequencer_program: Union[str, Sequence], **kwargs: Union[str, int]
    ) -> Dict[str, Any]:
        """Compiles the given sequencer program on the AWG Core.

        Warning:
            After uploading the sequencer program one needs to wait before for
            the awg core to become ready before it can be enabled.
            The awg core indicates the ready state through its `ready` node.
            (device.awgs[0].ready() == True)

        Args:
            sequencer_program: Sequencer program to be uploaded.

        Keyword Args:
            samplerate: Target sample rate of the sequencer. Only allowed/
                necessary for HDAWG devices. Must correspond to the samplerate
                used by the device (device.system.clocks.sampleclock.freq()).
                If not specified the function will get the value itself from
                the device. It is recommended passing the samplerate if more
                than one sequencer code is uploaded in a row to speed up the
                execution time.
            wavepath: path to directory with waveforms. Defaults to path used
                by LabOne UI or AWG Module.
            waveforms: waveform CSV files separated by ';'
            output: name of embedded ELF filename.

        Examples:
            >>> compile_info = device.awgs[0].load_sequencer_program(seqc)
            >>> device.awgs[0].ready.wait_for_state_change(1)
            >>> device.awgs[0].enable(True)

        Raises:
            RuntimeError: `sequencer_program` is empty.
            RuntimeError: If the upload or compilation failed.

        .. versionadded:: 0.3.4

            `sequencer_program` does not accept empty strings

        .. versionadded:: 0.4.0

            Use offline compiler instead of AWG module to compile the sequencer
            program. This speeds of the compilation and also enables parallel
            compilation/upload.
        """
        return self._tk_object.load_sequencer_program(
            sequencer_program=sequencer_program, **kwargs
        )

    def write_to_waveform_memory(
        self, waveforms: Waveforms, indexes: list = None
    ) -> None:
        """Writes waveforms to the waveform memory.

        The waveforms must already be assigned in the sequencer program.

        Args:
            waveforms: Waveforms that should be uploaded.
            indexes: Specify a list of indexes that should be uploaded. If
                nothing is specified all available indexes in waveforms will
                be uploaded. (default = None)

        .. versionchanged:: 0.4.2

            Removed `validate` flag and functionality. The validation check is
            now done in the `Waveforms.validate` function.
        """
        return self._tk_object.write_to_waveform_memory(
            waveforms=waveforms, indexes=indexes
        )

    def read_from_waveform_memory(self, indexes: List[int] = None) -> Waveforms:
        """Read waveforms from the waveform memory.

        Args:
            indexes: List of waveform indexes to read from the device. If not
                specified all assigned waveforms will be downloaded.

        Returns:
            Waveform object with the downloaded waveforms.
        """
        return self._tk_object.read_from_waveform_memory(indexes=indexes)


class Integration(ZINode):
    """Integration part for the UHFQA.

    Args:
        root: Underlying node tree.
        tree: tree (node path as tuple) of the corresponding node.

    .. versionadded:: 0.3.2
    """

    def __init__(self, parent, tk_object, snapshot_cache=None, zi_node=None):
        ZINode.__init__(
            self, parent, "integration", snapshot_cache=snapshot_cache, zi_node=zi_node
        )
        self._tk_object = tk_object

    def write_integration_weights(self, weights: Union[Waveforms, dict]) -> None:
        """Upload complex integration weights.

        The weight functions are applied to the real and imaginary part of
        the input signal. In the hardware the weights are implemented
        as 17-bit integers.

        Args:
            weights: Dictionary containing the weight functions, where
                keys correspond to the indices of the integration weights to be
                configured.

        Note:
            Does not raise an error when sample limit is exceeded, but applies only
            the maximum number of samples. Please refer to LabOne node documentation
            for the number of maximum integration weight samples.

        Note:
            This function calls both `/qas/n/integration/weights/n/real` and
            `/qas/n/integration/weights/n/imag` nodes.

            If only real or imaginary part is defined, the number of defined samples
            from the other one is zeroed.
        """
        return self._tk_object.write_integration_weights(weights=weights)


class QAS(ZINode):
    """Quantum Analyzer Channel for the UHFQA.

    Args:
        root: Underlying node tree.
        tree: tree (node path as tuple) of the corresponding node.
    """

    def __init__(self, parent, tk_object, index, snapshot_cache=None, zi_node=None):
        ZINode.__init__(
            self, parent, f"qas_{index}", snapshot_cache=snapshot_cache, zi_node=zi_node
        )
        self._tk_object = tk_object
        if self._tk_object.integration:

            self.add_submodule(
                "integration",
                Integration(
                    self,
                    self._tk_object.integration,
                    zi_node=self._tk_object.integration.node_info.path,
                    snapshot_cache=self._snapshot_cache,
                ),
            )

    def crosstalk_matrix(self, matrix: np.ndarray = None) -> Optional[np.ndarray]:
        """Sets or gets the crosstalk matrix of the UHFQA as a 2D array.

        Args:
            matrix: The 2D matrix used in the digital signal
                processing path to compensate for crosstalk between the
                different channels. The given matrix can also be a part
                of the entire 10 x 10 matrix. Its maximum dimensions
                are 10 x 10 (default: None).

        Returns:
            If no argument is given the method returns the current
            crosstalk matrix as a 2D numpy array.

        Raises:
            ValueError: If the matrix size exceeds the maximum size of
                10 x 10

        """
        return self._tk_object.crosstalk_matrix(matrix=matrix)

    def adjusted_delay(self, value: int = None) -> int:
        """Set or get the adjustment in the quantum analyzer delay.

        Adjusts the delay that defines the time at which the integration starts
        in relation to the trigger signal of the weighted integration units.

        Depending if the deskew matrix is bypassed there exists a different
        default delay. This function can be used to add an additional delay to
        the default delay.

        Args:
            value: Number of additional samples to adjust the delay. If not
                specified this function will just return the additional delay
                currently set.

        Returns:
            The adjustment in delay in units of samples.

        Raises:
            ValueError: If the adjusted quantum analyzer delay is outside the
                allowed range of 1021 samples.

        """
        return self._tk_object.adjusted_delay(value=value)


class UHFQA(ZIBaseInstrument):
    """QCoDeS driver for the Zurich Instruments UHFQA."""

    def _init_additional_nodes(self):
        """Init class specific modules and parameters."""
        if self._tk_object.awgs:

            channel_list = ZIChannelList(
                self,
                "awgs",
                AWG,
                zi_node=self._tk_object.awgs.node_info.path,
                snapshot_cache=self._snapshot_cache,
            )
            for i, x in enumerate(self._tk_object.awgs):
                channel_list.append(
                    AWG(
                        self,
                        x,
                        i,
                        zi_node=self._tk_object.awgs[i].node_info.path,
                        snapshot_cache=self._snapshot_cache,
                    )
                )
            # channel_list.lock()
            self.add_submodule("awgs", channel_list)

        if self._tk_object.qas:

            channel_list = ZIChannelList(
                self,
                "qas",
                QAS,
                zi_node=self._tk_object.qas.node_info.path,
                snapshot_cache=self._snapshot_cache,
            )
            for i, x in enumerate(self._tk_object.qas):
                channel_list.append(
                    QAS(
                        self,
                        x,
                        i,
                        zi_node=self._tk_object.qas[i].node_info.path,
                        snapshot_cache=self._snapshot_cache,
                    )
                )
            # channel_list.lock()
            self.add_submodule("qas", channel_list)

    def enable_qccs_mode(self) -> None:
        """Configure the instrument to work with PQSC.

        This method sets the reference clock source and DIO settings
        correctly to connect the instrument to the PQSC.

        Info:
            Use ``factory_reset`` to reset the changes if necessary
        """
        return self._tk_object.enable_qccs_mode()



"""Autogenerated module for the SweeperModule QCoDeS driver."""
import typing as t
from zhinst.toolkit.driver.modules.sweeper_module import (
    SweeperModule as TKSweeperModule,
)

from zhinst.qcodes.driver.modules.base_module import ZIBaseModule


if t.TYPE_CHECKING:
    from zhinst.qcodes.session import Session


class ZISweeperModule(ZIBaseModule):
    """Implements a base Sweeper Module for Lock-In instruments.

    The Sweeper Module allows for simple and efficient parameter sweeps while
    acquiring data streams from multiple different signal sources. The module
    supports well defined sweeps of various parameters as well as application
    specific measurement presets. For more information on how to use the Sweeper
    Module, have a look at the LabOne Programming Manual.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/sweeper_module.html


    Args:
        tk_object: Underlying zhinst-toolkit object.
        session: Session to the Data Server.
        name: Name of the module in QCoDeS.
    """

    def __init__(
        self,
        tk_object: TKSweeperModule,
        session: "Session",
        name: str = "sweeper_module",
    ):
        super().__init__(tk_object, session, name)

        self._tk_object.root.update_nodes(
            {
                "/gridnode": {
                    "GetParser": lambda value: self._get_node(value),
                    "SetParser": lambda value: self._set_node(value),
                }
            },
            raise_for_invalid_node=False,
        )

    def finish(self) -> None:
        """Stop the module.

        .. versionadded:: 0.5.0
        """
        return self._tk_object.finish()



"""Autogenerate the QCoDeS drivers from toolkit and zhinst-core."""
from collections import namedtuple
import typing as t
import inspect
import re
import importlib
import jinja2
import isort
import black
import autoflake
import click
from pathlib import Path

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.driver.modules.base_module import BaseModule
from zhinst.toolkit.nodetree import Node, NodeTree
import conf

parameter_tuple = namedtuple("parameter", ["name", "is_node"])
submodule_tuple = namedtuple("submodule", ["subclass", "name", "is_list"])
function_tuple = namedtuple("function", ["name", "is_deprecated"])
class_tuple = namedtuple("toolkit_class", ["functions", "parameters", "sub_modules"])


def getPropertyInfo(
    name: str, property: object, class_type: object
) -> t.Union[parameter_tuple, submodule_tuple]:
    """Get all necessary information for a toolkit property.

    Args:
        name (str): Name of the property
        property (object): Class of the property
        class_type (object): Owner class of the property

    Returns:
        Union[parameter_tuple,submodule_tuple]
    """
    typehint = t.get_type_hints(property.fget)
    if "deprecated" in inspect.getsource(property.fget):
        # TODO decide if we should keep them or remove them?
        print(f"WARNING {name}: deprecated property -> ignored")
    elif "typing.Union" in str(typehint["return"]) or "typing.Sequence" in str(
        typehint["return"]
    ):
        try:
            item_class = typehint["return"].__args__[0]
            if "typing.Sequence" in str(item_class):
                return submodule_tuple(item_class.__args__[0], name, True)
            if issubclass(item_class, Node):
                # list of submodules
                return submodule_tuple(item_class, name, True)
        except Exception:
            raise Exception(
                f"{name} in {class_type} has a type hint list, "
                "but misses the item typehint"
            )
        if item_class.__name__ in ["str"]:
            return parameter_tuple(name, False)
        raise Exception(
            f"{name} in {class_type} has a type hint list, "
            f"but the item typehint is of {str(item_class)}."
            "only submodules are currently allowed."
        )
    elif hasattr(typehint["return"], "_name") and typehint["return"]._name == "List":
        return parameter_tuple(name, False)
    elif typehint["return"] == Node:
        return parameter_tuple(name, True)
    elif typehint["return"] == NodeTree:
        raise RuntimeError("Why would there be such thing?")
    elif (
        inspect.isclass(typehint["return"])
        and issubclass(typehint["return"], Node)
        or (
            "typing.Optional" in str(typehint["return"])
            and issubclass(typehint["return"].__args__[0], Node)
        )
    ):
        return submodule_tuple(typehint["return"], name, False)
    return parameter_tuple(name, False)


def getInfo(class_type: object, existing_names: list) -> t.Tuple[class_tuple, list]:
    """Get all necessary information for a toolkit class.

    Args:
        class_type (object) toolkit class
        existing_names (list) list of existing names for the QCoDeS class.

    Returns:
        class_tuple: functions,parameters,sub_modules
        list:        updated existing names
    """
    blacklist_names = []
    functions = []
    parameters = []
    sub_modules = []

    for name, attribute in vars(class_type).items():
        # ignore private/blacklisted and existing items
        if name.startswith("_") or name in blacklist_names or name in existing_names:
            continue
        if isinstance(attribute, property):
            property_info = getPropertyInfo(name, attribute, class_type)
            if isinstance(property_info, parameter_tuple):
                parameters.append(property_info)
            else:
                sub_modules.append(property_info)
        else:
            # function
            if not callable(attribute):
                raise RuntimeError("Unsupported class item")
            functions.append(
                function_tuple(name, "deprecated" in inspect.getsource(attribute))
            )
        existing_names.append(name)

    result_tuple = class_tuple(functions, parameters, sub_modules)

    # add base class info if it is inside toolkit
    blacklist_base = [
        Node,
        BaseInstrument,
        BaseModule,
    ]
    for base_class in class_type.__bases__:
        if (
            base_class not in blacklist_base
            and "zhinst.toolkit" in base_class.__module__
        ):
            base_info, existing_names = getInfo(base_class, existing_names)
            result_tuple = class_tuple(
                *list(
                    base_tuple + res_tuple
                    for base_tuple, res_tuple in zip(base_info, result_tuple)
                )
            )

    return result_tuple, existing_names


def generate_submodules_info(sub_modules: list, base_class: object) -> t.List[list]:
    """Gather information for the submodules.

    Args:
        sub_modules(list) list of classes that should be add as submodules.

    Returns:
        (list) infos for the parent class
        (list) newly generate class information
    """
    parent_info = []  # information for the parent class
    submodule_info = []  # information about the submodule classes
    for module, name, is_list in sub_modules:
        if name == "commandtable" and base_class.__name__ == "Generator":
            continue
        if "typing.Optional" in str(module):
            module = module.__args__[0]
        parent_info.append(
            {"name": name, "class_name": module.__name__, "is_list": is_list}
        )
        submodule_info += generate_qcodes_class_info(
            module,
            is_submodule=True,
            is_list_element=is_list,
        )
    return parent_info, submodule_info


def generate_parameter_info(parameters: list, class_type: object) -> list:
    """Gather information for the Qcodes parameter.

    Args:
        parameters (list): list of parameters that should be added.
        toolkit_class (object): toolkit class that the parameters belong to.

    Returns:
        (list) list dict with information for each parameter.
    """
    parameter_info = []
    has_node_param = False
    for parameter, is_node in parameters:
        has_node_param = True if has_node_param or is_node else False
        signature = inspect.signature(getattr(class_type, parameter).fget)
        try:
            return_annotation = signature.return_annotation.__name__
        except AttributeError:
            return_annotation = str(signature.return_annotation)
            return_annotation = return_annotation.replace("typing.", "t.")

        parameter_info.append(
            {
                "name": parameter,
                "is_node": is_node,
                "docstring": getattr(class_type, parameter).__doc__,
                "return_annotation": return_annotation,
            }
        )
    return parameter_info, has_node_param


def generate_functions_info(functions: list, toolkit_class: object) -> list:
    """Gather information for the functions.

    Args:
        functions (list): list of functions that should be added.
        toolkit_class (object): toolkit class that the functions belong to.

    Returns:
        (list) list dict with information for each function.
    """
    # Enums from toolkit should be exposed also in the QCoDeS driver
    enums = {
        "NodeDict": "zhinst.toolkit.nodetree.helper.NodeDict",
        "np.": "numpy.",
        "QuditSettings": "zhinst.utils.shfqa.multistate.QuditSettings",
        "SHFQAChannelMode": "zhinst.toolkit.interface.SHFQAChannelMode",
        "MappingMode": "zhinst.toolkit.interface.MappingMode",
        "TriggerImpedance": "zhinst.toolkit.interface.TriggerImpedance",
        "AveragingMode": "zhinst.toolkit.interface.AveragingMode",
        "Waveforms": "zhinst.toolkit.waveform.Waveforms",
        "CommandTable": "zhinst.toolkit.command_table.CommandTable",
        "Sequence": "zhinst.toolkit.sequence.Sequence",
        "Path": "pathlib.Path",
        '"DeviceType"': "ForwardRef('DeviceType')",
    }
    # regex to find deprecation decorator
    deprecated_regex = re.compile(r"(@depreca(.|\n)*?)\s*?(?:def|@)")

    functions_info = []
    for name, deprecated in functions:
        decorator = ""
        if deprecated:
            # copy deprecation decorator
            deprecation_deco = inspect.getsource(getattr(toolkit_class, name))
            decorator = deprecated_regex.search(deprecation_deco).group(1)
        docstring = getattr(toolkit_class, name).__doc__
        signature = inspect.signature(getattr(toolkit_class, name))
        signature_str = str(signature)
        is_node_doc = False
        # replace toolkit enum typehint with direct typehint
        for new, old in enums.items():
            signature_str = signature_str.replace(old, new)
            if new in signature_str:
                # search for enum typehint in function arguments
                regex_result = re.search(
                    f"<({new}.*?):.*?>",
                    signature_str,
                )
                if regex_result:
                    signature_str = signature_str.replace(
                        regex_result.group(0), regex_result.group(1)
                    )
                if new == "NodeDict":
                    is_node_doc = True

        # replace weird stuff
        for weird_stuff, replacement in conf.TYPE_HINT_REPLACEMENTS.items():
            signature_str = signature_str.replace(weird_stuff, replacement)

        call_parameter_str = []
        for param in signature.parameters:
            if param == "self":
                continue
            if param == "kwargs":
                call_parameter_str.append("**kwargs")
            elif param == "args":
                call_parameter_str.append("*args")
            else:
                call_parameter_str.append(f"{param}={param}")
        call_signature = ", ".join(call_parameter_str)

        functions_info.append(
            {
                "name": name,
                "decorator": decorator,
                "signature": signature_str,
                "docstring": docstring if docstring else "",
                "call_signature": call_signature,
                "return_annotation": str(signature.return_annotation)
                if signature.return_annotation
                else "",
                "is_node_dict": is_node_doc,
            }
        )
    return functions_info


def generate_qcodes_class_info(
    class_type: object,
    is_submodule: bool = False,
    is_list_element: bool = False,
):
    """Gather information for a qcodes class from a toolkit class.

    Args:
        class_type (object): toolkit class that should mimiced.
        is_submodule (bool): Flag if the class is a sub module of another
            toolkit class.
        is_list_element (bool): Flag if the class is instaciated in a list.

    Returns:
        (list) list of dicts with information for the class and its subclasses.
    """
    tk_class_info, _ = getInfo(class_type, [])
    submodule_info, gathered_info = generate_submodules_info(
        tk_class_info.sub_modules, class_type
    )
    parameter_info, has_node_param = generate_parameter_info(
        tk_class_info.parameters, class_type
    )
    function_info = generate_functions_info(tk_class_info.functions, class_type)

    display_name_replacements = {"CommandTableNode": "commandtable", "AWGCore": "awg"}

    gathered_info.append(
        {
            "name": class_type.__name__,
            "display_name": display_name_replacements.get(
                class_type.__name__, class_type.__name__.lower()
            ),
            "is_instrument_class": not is_submodule,
            "docstring": class_type.__doc__,
            "modules": submodule_info,
            "functions": function_info,
            "parameters": parameter_info,
            "has_node_param": has_node_param,
            "is_list": is_list_element,
        }
    )
    return gathered_info


def camel_to_snake(name: str) -> str:
    """Convert camelcase into snake case.

    Args:
        name  (str): name in camel case

    Returns:
        str: name in snake case
    """
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def generate_qcodes_driver(
    toolkit_class: object,
    template_path: t.Union[str, Path] = conf.TEMPLATE_PATH,
    output_dir: t.Union[str, Path] = conf.OUTPUT_DIR_DEVICES_DRIVER,
) -> None:
    """Generates the Qcodes drivers for the toolkit instrument classes.

    Args:
        toolkit_class (object): toolkit instrument driver class
        template_path (str):  jinja template location (default = "templates/")
        output_dir (str): output directory (default = "src/zhinst/qcodes/modules/")
    """
    data = {
        "classes": generate_qcodes_class_info(toolkit_class),
        "name": toolkit_class.__name__,
    }

    templateLoader = jinja2.FileSystemLoader(searchpath=template_path)
    templateEnv = jinja2.Environment(loader=templateLoader)
    template = templateEnv.get_template("instrument_class.py.j2")
    result = template.render(data)
    # result = black.format_str(result, mode=black.FileMode())
    result = black.format_str(
        result,
        mode=black.mode.Mode(
            target_versions={black.TargetVersion.PY310},
            line_length=88,
        ),
    )
    result = autoflake.fix_code(result, remove_all_unused_imports=True)
    module_name = camel_to_snake(toolkit_class.__name__)
    py_filename = str(output_dir) + "/" + module_name.lower() + ".py"
    with open(py_filename, "w+") as outfile:
        outfile.write(result)
    print(f"{py_filename} created.")


def generate_qcodes_driver_modules(
    zi_module_class: object,
    template_path: t.Union[str, Path] = conf.TEMPLATE_PATH,
    output_dir: t.Union[str, Path] = conf.OUTPUT_DIR_MODULE_DRIVER,
) -> None:
    """Generates the Qcodes drivers for the ziPython modules.

    The drivers forward all function (except the one in black_list) and
    initialises the nodes a nested qcodes prameter.

    Args:
        zi_module_class (object): ziPython module class
        template_path (str):  jinja template location (default = "templates/")
        output_dir (str): output directory (default = "src/zhinst/qcodes/modules/")
    """
    ignored_functions = []
    if zi_module_class == BaseModule:
        ignored_functions.append("subscribe")
        ignored_functions.append("unsubscribe")
    tk_class_info, _ = getInfo(zi_module_class, [])
    qcodes_functions = []
    for function in tk_class_info.functions:
        if function.name not in ignored_functions:
            qcodes_functions.append(function)

    function_info = generate_functions_info(qcodes_functions, zi_module_class)
    name = zi_module_class.__name__
    module_name = camel_to_snake(zi_module_class.__name__)
    module_docstring = zi_module_class.__doc__.split("Args:")[0]
    node_param = []
    if module_name == "sweeper_module":
        node_param.append("gridnode")
    if module_name == "daq_module":
        node_param.append("triggernode")
    data = {
        "name": name,
        "module_name": module_name,
        "module_docstring": module_docstring,
        "base_module": "ZIInstrument"
        if zi_module_class == BaseModule
        else "ZIBaseModule",
        "functions": function_info,
        "node_param": node_param,
    }
    templateLoader = jinja2.FileSystemLoader(searchpath=template_path)
    templateEnv = jinja2.Environment(loader=templateLoader)
    template = templateEnv.get_template("module_class.py.j2")
    result = template.render(data)
    result = black.format_str(
        result,
        mode=black.mode.Mode(
            target_versions={black.TargetVersion.PY310},
            line_length=88,
        ),
    )
    result = autoflake.fix_code(result, remove_all_unused_imports=True)
    py_filename = str(output_dir) + "/" + module_name.lower() + ".py"
    with open(py_filename, "w+") as outfile:
        outfile.write(result)
    print(f"{py_filename} created.")


def generate_device_api():
    """Generate device API.

    The device API enables that every ZI device is exposed as a class that
    can be istanciated directly without a session.
    """
    DEVICE_API_FILEPATH = "src/zhinst/qcodes/device_creator.py"
    data = {
        "classes": [
            {"name": "ZIDevice", "parent": "ZIBaseInstrument", "is_hf2": False},
            {"name": "SHFQA", "parent": "SHFQADriver", "is_hf2": False},
            {"name": "SHFSG", "parent": "SHFSGDriver", "is_hf2": False},
            {"name": "HDAWG", "parent": "HDAWGDriver", "is_hf2": False},
            {"name": "PQSC", "parent": "PQSCDriver", "is_hf2": False},
            {"name": "SHFQC", "parent": "SHFQCDriver", "is_hf2": False},
            {"name": "UHFLI", "parent": "UHFLIDriver", "is_hf2": False},
            {"name": "UHFQA", "parent": "UHFQADriver", "is_hf2": False},
            {"name": "MFLI", "parent": "ZIBaseInstrument", "is_hf2": False},
            {"name": "MFIA", "parent": "ZIBaseInstrument", "is_hf2": False},
            {"name": "HF2", "parent": "ZIBaseInstrument", "is_hf2": True},
        ],
        "imports": [
            "from zhinst.qcodes.driver.devices.base import ZIBaseInstrument",
            "from zhinst.qcodes.driver.devices.hdawg import HDAWG as HDAWGDriver",
            "from zhinst.qcodes.driver.devices.pqsc import PQSC as PQSCDriver",
            "from zhinst.qcodes.driver.devices.shfqa import SHFQA as SHFQADriver",
            "from zhinst.qcodes.driver.devices.shfqc import SHFQC as SHFQCDriver",
            "from zhinst.qcodes.driver.devices.shfsg import SHFSG as SHFSGDriver",
            "from zhinst.qcodes.driver.devices.uhfli import UHFLI as UHFLIDriver",
            "from zhinst.qcodes.driver.devices.uhfqa import UHFQA as UHFQADriver",
        ],
    }
    templateLoader = jinja2.FileSystemLoader(searchpath=conf.TEMPLATE_PATH)
    templateEnv = jinja2.Environment(loader=templateLoader)
    template = templateEnv.get_template("device_api.py.j2")
    result = template.render(data)
    result = black.format_str(
        result,
        mode=black.mode.Mode(
            target_versions={black.TargetVersion.PY37},
            line_length=88,
            string_normalization=False,
        ),
    )
    result = autoflake.fix_code(result, remove_all_unused_imports=True)
    result = isort.code(result)
    with open(DEVICE_API_FILEPATH, "w+") as outfile:
        outfile.write(result)
    print(f"{DEVICE_API_FILEPATH} created.")


@click.group()
def main():
    """Autogeneration of QCoDeS driver from zhinst.toolkit."""


@main.command(help="toolkit instrument class")
@click.argument(
    "name",
    required=False,
    type=str,
)
def instrument_class(name):
    """Generate a insturment class."""
    module = importlib.import_module(f"{conf.TOOLKIT_DEVICE_MODULE}.{name.lower()}")
    generate_qcodes_driver(getattr(module, name.upper()))


@main.command(help="toolkit module class")
@click.argument(
    "name",
    required=False,
    type=str,
)
def module_class(name):
    """Generate a module class."""
    module = importlib.import_module(f"{conf.TOOLKIT_MODULE_MODULE}.{name}")
    generate_qcodes_driver_modules(getattr(module, name.upper()))


@main.command(help="Generate all.")
def generate_all():
    """Generate all drivers."""
    for name in conf.DEVICE_DRIVERS:
        module = importlib.import_module(f"{conf.TOOLKIT_DEVICE_MODULE}.{name.lower()}")
        generate_qcodes_driver(getattr(module, name.upper()))
    generate_device_api()

    for name in conf.MODULE_DRIVERS:
        module = importlib.import_module(
            f"{conf.TOOLKIT_MODULE_MODULE}.{camel_to_snake(name)}"
        )
        generate_qcodes_driver_modules(getattr(module, name))


if __name__ == "__main__":
    main()



"""Autogenerated module for the PIDAdvisorModule QCoDeS driver."""
import typing as t
from zhinst.toolkit.driver.modules.pid_advisor_module import (
    PIDAdvisorModule as TKPIDAdvisorModule,
)

from zhinst.qcodes.driver.modules.base_module import ZIBaseModule


if t.TYPE_CHECKING:
    from zhinst.qcodes.session import Session


class ZIPIDAdvisorModule(ZIBaseModule):
    """PID Advisor Module.

    The PID Advisor Module provides the functionality available in the Advisor,
    Tuner and Display sub-tabs of the LabOne User Interface’s PID / PLL tab.
    The PID Advisor is a mathematical model of the instrument’s PID and can be
    used to calculate PID controller parameters for optimal feedback loop
    performance. The controller gains calculated by the module can be easily
    transferred to the device via the API and the results of the Advisor’s
    modeling are available as Bode and step response plot data.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/pid_advisor_module.html


    Args:
        tk_object: Underlying zhinst-toolkit object.
        session: Session to the Data Server.
        name: Name of the module in QCoDeS.
    """

    def __init__(
        self,
        tk_object: TKPIDAdvisorModule,
        session: "Session",
        name: str = "pid_advisor_module",
    ):
        super().__init__(tk_object, session, name)

    def wait_done(self, *, timeout: float = 20.0, sleep_time: float = 2) -> None:
        """Waits until the pid advisor is finished.

        Args:
            timeout (float): The maximum waiting time in seconds for the
                measurement (default: 20).
            sleep_time (int): Time in seconds to wait between
                requesting sweeper state. (default: 0.5)

        Raises:
            TimeoutError: The measurement is not completed before
                timeout.
        """
        return self._tk_object.wait_done(timeout=timeout, sleep_time=sleep_time)



"""Autogenerated module for the PrecompensationAdvisorModule QCoDeS driver."""
import typing as t
from zhinst.toolkit.driver.modules.precompensation_advisor_module import (
    PrecompensationAdvisorModule as TKPrecompensationAdvisorModule,
)

from zhinst.qcodes.driver.modules.base_module import ZIBaseModule


if t.TYPE_CHECKING:
    from zhinst.qcodes.session import Session


class ZIPrecompensationAdvisorModule(ZIBaseModule):
    """Precompensation Advisor Module.

    This module provides the functionality available in the LabOne User
    Interface’s Precompensation Tab. In essence the precompensation allows a
    pre-distortion or pre-emphasis to be applied to a signal before it leaves
    the instrument, to compensate for undesired distortions caused by the
    device under test (DUT). The Precompensation Advisor module simulates the
    precompensation filters in the device, allowing the user to experiment with
    different filter settings and filter combinations to obtain an optimal
    output signal, before using the setup in the actual device.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/precompensation_advisor_module.html

    Note:
        Unlike most other LabOne modules, this module does not expose any
        functions. Each time one or more filter parameters are changed, the
        module re-runs the simulation and the results can be read via the
        wave/output, wave/output/forwardwave and wave/output/backwardwave
        parameters.


    Args:
        tk_object: Underlying zhinst-toolkit object.
        session: Session to the Data Server.
        name: Name of the module in QCoDeS.
    """

    def __init__(
        self,
        tk_object: TKPrecompensationAdvisorModule,
        session: "Session",
        name: str = "precompensation_advisor_module",
    ):
        super().__init__(tk_object, session, name)



"""Autogenerated module for the DAQModule QCoDeS driver."""
import typing as t
from zhinst.toolkit.driver.modules.daq_module import DAQModule as TKDAQModule

from zhinst.qcodes.driver.modules.base_module import ZIBaseModule

from zhinst.qcodes.qcodes_adaptions import (
    NodeDict,
)

if t.TYPE_CHECKING:
    from zhinst.qcodes.session import Session


class ZIDAQModule(ZIBaseModule):
    """Data Acquisition Module.

    The Data Acquisition Module corresponds to the Data Acquisition tab of the
    LabOne User Interface. It enables the user to record and align time and
    frequency domain data from multiple instrument signal sources at a defined
    data rate. The data may be recorded either continuously or in bursts based
    upon trigger criteria analogous to the functionality provided by laboratory
    oscilloscopes.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/data_acquisition_module.html


    Args:
        tk_object: Underlying zhinst-toolkit object.
        session: Session to the Data Server.
        name: Name of the module in QCoDeS.
    """

    def __init__(
        self, tk_object: TKDAQModule, session: "Session", name: str = "daq_module"
    ):
        super().__init__(tk_object, session, name)

        self._tk_object.root.update_nodes(
            {
                "/triggernode": {
                    "GetParser": lambda value: self._get_node(value),
                    "SetParser": lambda value: self._set_node(value),
                }
            },
            raise_for_invalid_node=False,
        )

    def finish(self) -> None:
        """Stop the module.

        .. versionadded:: 0.5.0
        """
        return self._tk_object.finish()

    def finished(self) -> bool:
        """Check if the acquisition has finished.

        Returns:
            Flag if the acquisition has finished.

        .. versionadded:: 0.5.0
        """
        return self._tk_object.finished()

    def trigger(self) -> None:
        """Execute a manual trigger.

        .. versionadded:: 0.5.0
        """
        return self._tk_object.trigger()

    def read(self, *, raw: bool = False, clk_rate: float = 60000000.0) -> NodeDict:
        """Read the acquired data from the module.

        The data is split into bursts.

        Args:
            raw: Flag if the acquired data from the subscribed device
                device nodes should be converted into the DAQResult format
                (raw = False) or not. (default = False)
            clk_rate: Clock rate [Hz] for converting the timestamps. Only
                applies if the raw flag is reset.

        Returns:
            Result of the burst grouped by the signals.
        """
        return NodeDict(self._tk_object.read(raw=raw, clk_rate=clk_rate))



"""Toolkit adaption for the zhinst.utils.SHFSweeper."""
import typing as t

from zhinst.toolkit.driver.modules.shfqa_sweeper import SHFQASweeper as TKSHFQASweeper

from zhinst.qcodes.qcodes_adaptions import init_nodetree, ZIInstrument

if t.TYPE_CHECKING:
    from zhinst.qcodes.driver.devices import DeviceType
    from zhinst.qcodes.session import Session


class ZISHFQASweeper(ZIInstrument):
    """QCoDeS adaption for the zhinst.utils.SHFSweeper.

    For now the general sweeper module does not support the SHFQA. However a
    python based implementation called ``SHFSweeper`` does already provide
    this functionality. The ``SHFSweeper`` is part of the ``zhinst`` module
    and can be found in the utils.

    Toolkit wraps around the ``SHFSweeper`` and exposes a interface that is
    similar to the LabOne modules, meaning the parameters are exposed in a
    node tree like structure.

    All parameters can be accessed through their corresponding node:
    * device: Device to run the sweeper with
    * sweep: Frequency range settings for a sweep
    * rf: RF in- and output settings for a sweep
    * average: Averaging settings for a sweep
    * trigger: Settings for the trigger
    * envelope: Settings for defining a complex envelope for pulsed spectroscopy

    The underlying module is updated with the parameter changes automatically.
    Every functions from the underlying SHFSweeper module is exposed and can be
    used in the same way.

    Args:
        tk_object: Instance of the toolkit shfqa sweeper.
        session: Session to the Data Server.
    """

    def __init__(self, tk_object: TKSHFQASweeper, session: "Session"):
        super().__init__(
            f"zi_shfqasweeper_{len(self.instances())}", tk_object.root, is_module=True
        )
        self._tk_object = tk_object
        self._session = session
        init_nodetree(self, self._tk_object, self._snapshot_cache)
        self._tk_object.root.update_nodes(
            {"/device": {"GetParser": lambda value: self._get_device(value)}}
        )

    def _get_device(self, serial: str) -> t.Union["DeviceType", str]:
        """Convert a device serial into a QCoDeS device object.

        Args:
            serial: Serial of the device

        Returns:
            QCoDeS device object. If the serial does not
                match to a connected device the serial is returned instead.
        """
        try:
            return self._session.devices[serial]
        except (RuntimeError, KeyError):
            return serial

    def run(self) -> dict:
        """Perform a sweep with the specified settings.

        This method eventually wraps around the `run` method of
        `zhinst.utils.shf_sweeper`

        Returns:
             A dictionary with measurement data of the last sweep.
        """
        return self._tk_object.run()

    def get_result(self) -> dict:
        """Get the measurement data of the last sweep.

        This method eventually wraps around the `get_result` method of
        `zhinst.utils.shf_sweeper`

        Returns:
             A dictionary with measurement data of the last sweep.

        """
        return self._tk_object.get_result()

    def plot(self) -> None:
        """Plot power over frequency for last sweep.

        This method eventually wraps around the `plot` method of
        `zhinst.utils.shf_sweeper`
        """
        return self._tk_object.plot()

    def get_offset_freq_vector(self) -> t.Any:
        """Get vector of frequency points.

        This method wraps around the `get_offset_freq_vector` method of
        `zhinst.utils.shf_sweeper`

        Returns:
            Vector of frequency points.
        """
        return self._tk_object.get_offset_freq_vector()



"""Base modules for the Zurich Instrument specific QCoDeS driver."""
import typing as t

from zhinst.toolkit.driver.devices import DeviceType

from zhinst.qcodes.qcodes_adaptions import init_nodetree, ZIInstrument

if t.TYPE_CHECKING:
    from zhinst.qcodes.session import ZISession, Session
    from qcodes.instrument.base import Instrument


class ZIBaseInstrument(ZIInstrument):
    """Generic QCoDeS driver for a Zurich Instrument device.

    All device specific class are derived from this class.
    It implements common functions valid for all
    devices.
    It also can be used directly, e.g. for instrument types that have no special
    class in QCoDeS.

    Args:
        tk_object: Instance of the toolkit base instrument
        name: Name of the instrument in qcodes. (default = "zi_{dev_type}_{serial}")
        raw: Flag if qcodes instance should only created with the nodes and not
            forwarding the toolkit functions. (default = False)
    """

    def __init__(
        self,
        tk_object: DeviceType,
        session: t.Union["ZISession", "Session", "Instrument"],
        name: t.Optional[str] = None,
        raw: bool = False,
    ):
        self._tk_object = tk_object
        self._session = session
        if not name:
            name = (
                f"zi_{tk_object.__class__.__name__.lower()}_{tk_object.serial.lower()}"
            )
        super().__init__(name, self._tk_object.root)

        if not raw:
            self._init_additional_nodes()
        init_nodetree(self, self._tk_object.root, self._snapshot_cache)

    def get_idn(self) -> t.Dict[str, t.Optional[str]]:
        """Fake a standard VISA ``*IDN?`` response."""
        return {
            "vendor": "Zurich Instruments",
            "model": self.device_type,
            "serial": self.serial,
            "firmware": self.system.fwrevision(),
        }

    def _init_additional_nodes(self) -> None:
        """Init additional qcodes parameter."""

    def factory_reset(self, deep: bool = True) -> None:
        """Load the factory default settings.

        Arguments:
            deep (bool): A flag that specifies if a synchronization
                should be performed between the device and the data
                server after loading the factory preset (default: True).
        """
        return self._tk_object.factory_reset(deep=deep)

    def check_compatibility(self) -> None:
        """Check if the software stack is compatible.

        Only if all versions and revisions of the software stack match stability
        can be ensured. The following criteria are checked:

        * minimum required zhinst-utils package is installed
        * minimum required zhinst-core package is installed
        * zhinst package matches the LabOne Data Server version
        * firmware revision matches the LabOne Data Server version

        Raises:
            ConnectionError: If the device is currently updating
            RuntimeError: If one of the above mentioned criteria is not
                fulfilled
        """
        self._tk_object.check_compatibility

    def get_streamingnodes(self) -> list:
        """Create a dictionary with all streaming nodes available."""
        return self._tk_object.get_streamingnodes()

    def set_transaction(self):
        """Context manager for a transactional set.

        Can be used as a context in a with statement and bundles all node set
        commands into a single transaction. This reduces the network overhead
        and often increases the speed.

        Within the with block a set commands to a node will be buffered
        and bundled into a single command at the end automatically.
        (All other operations, e.g. getting the value of a node, will not be
        affected)

        Warning:
            The set is always performed as deep set if called on device nodes.

        Examples:
            >>> with device.set_transaction():
                    device.test[0].a(1)
                    device.test[1].a(2)
        """
        return self._tk_object.set_transaction()

    @property
    def serial(self) -> str:
        """Instrument specific serial."""
        return self._tk_object.serial

    @property
    def device_type(self) -> str:
        """Type of the instrument (e.g. MFLI)."""
        return self._tk_object.device_type

    def device_options(self) -> str:
        """Enabled options of the instrument."""
        return self._tk_object.device_options

    @property
    def session(self) -> "Session":
        """Underlying session the device is connected through."""
        return self._session



"""Base modules for the Zurich Instrument specific QCoDeS driver."""

import re
from datetime import datetime
import typing as t
from contextlib import contextmanager, nullcontext
from collections.abc import Mapping

import numpy as np
from qcodes.instrument.base import Instrument
from qcodes.instrument.channel import ChannelList, InstrumentChannel
from qcodes.instrument.parameter import Parameter
from qcodes.utils.validators import ComplexNumbers
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import NodeDict as TKNodeDict
from zhinst.toolkit.nodetree.node import NodeInfo


class ZISnapshotHelper:
    """Helper class for the snapshot with Zurich Instrument devices.

    Instead of getting each node with a single get command this class bundles
    the get into a single command and stores the returned values into a
    temporary dictionary.
    """

    def __init__(self, nodetree: NodeTree, is_module: bool = False):
        self._is_running = False
        self._value_dict: t.Dict[str, t.Any] = {}
        self._start = datetime.now()
        self._nodetree = nodetree
        self._is_module = is_module

    @contextmanager
    def snapshot(self, name: t.Optional[str] = None):
        """Context manager for a optimized snapshot with ZI devices."""
        is_owner = not self._is_running
        if is_owner:
            self._start_snapshot(name)
        try:
            yield
        finally:
            if is_owner:
                self._stop_snapshot()

    def _start_snapshot(self, name: t.Optional[str] = None) -> bool:
        """Start a snapshot and make a single get to the device.

        Args:
            name: Name of the subnode which the snapshot should
                be taken. If not specified a snapshot of all nodes will be taken.
                (default = None)

        Returns:
            bool: Flag if a new snapshot was started.
        """
        if not self._nodetree or self._is_running:
            return False
        self._is_running = True
        if not self._is_module:
            kwargs = {
                "excludestreaming": True,
                "settingsonly": False,
                "excludevectors": True,
                "flat": True,
            }
        else:
            kwargs = {"flat": True}
        prefix = self._nodetree.prefix_hide
        if not name:
            name = prefix if prefix else ""
        else:
            name = "/" + prefix + "/" + name
        self._value_dict = self._nodetree.connection.get(f"{name}/*", **kwargs)
        self._start = datetime.now()
        return True

    def _stop_snapshot(self) -> None:
        """Stop a snapshot to prevent use of outdate data by accident."""
        self._is_running = False
        self._value_dict = {}

    def get(self, parameter: Parameter, fallback_get: t.Callable) -> t.Any:
        """Get the value for a specific QCoDeS Parameter.

        Tries to mimic the behavior of a normal get (e.g. update cache).
        If the value is not found in the dictionary the fallback get is called.
        The fallback get should get the value from the device.

        Args:
            parameter: QCoDeS Parameter object
            fallback_get: fallback function to get the value from the device
        Returns:
            Value for the Node
        """
        value = self._value_dict.get(parameter.zi_node.lower())
        if value is not None:
            try:
                value = value["value"][0]
            except (IndexError, TypeError):
                # HF2 has no timestamp -> no dict
                value = value[0]
            # convert numpy types to standart types
            value = value.item() if hasattr(value, "item") else value
            # convert complex into string
            value = str(value) if isinstance(value, complex) else value
            parameter.cache._update_with(
                value=value, raw_value=value, timestamp=self._start
            )
        else:  # fallback is normal get
            value = fallback_get()
        return value

    @staticmethod
    def print_readable_snapshot(
        qcodes_object: Instrument, update: bool = False, max_chars: int = 80
    ) -> None:
        """Prints a readable version of the snapshot.

        The readable snapshot includes the name, value and unit of each
        parameter.
        A convenience function to quickly get an overview of the
        status of an instrument.

        Args:
            qcodes_object (object): Object for which the snapshot should be printed.
            update (bool): Flag if the state should be queried from the
                           instrument.
            max_chars (int): The maximum number of characters per line. The
                readable snapshot will be cropped if this value is exceeded.
                Defaults to 80 to be consistent with default terminal width.
        """
        floating_types = (float, np.integer, np.floating)
        snapshot = qcodes_object.snapshot(update=update)
        snapshot_parameters = snapshot.get("parameters")

        if snapshot_parameters:
            # Min of 50 is to prevent a super long parameter name to break this
            # function
            par_lengths = [len(p) for p in snapshot_parameters]
            par_field_len = min(max(par_lengths) + 1, 50) if par_lengths else 0

            print(qcodes_object.name + ":")
            print(f"\t{'parameter':<{par_field_len}}: value")
            print("\t" + "-" * (max_chars - 8))
            for parameter in sorted(snapshot_parameters):
                parameter = snapshot_parameters[parameter]
                name = parameter["name"]
                msg = f"\t{name:<{par_field_len}}:"

                # in case of e.g. ArrayParameters, that usually have
                # snapshot_value == False, the parameter may not have
                # a value in the snapshot
                val = parameter.get("value", "Not available")

                unit = parameter.get("unit", None)
                if unit is None:
                    # this may be a multi parameter
                    unit = parameter.get("units", None)
                if isinstance(val, floating_types):
                    msg += f"\t{val:.5g} "
                    # numpy float and int types format like builtins
                else:
                    msg += f"\t{val} "
                if unit != "":  # corresponds to no unit
                    msg += f"({unit})"
                # Truncate the message if it is longer than max length
                if len(msg) > max_chars and max_chars != -1:
                    msg = msg[0 : max_chars - 3] + "..."  # noqa: E203
                print(msg)

        for submodule in qcodes_object.submodules.values():
            submodule.print_readable_snapshot(update=update, max_chars=max_chars)

    @property
    def is_running(self) -> bool:
        """Flag if a snapshot is in progress."""
        return self._is_running


class ZIParameter(Parameter):
    """Zurich Instrument specific QCoDeS Parameter.

    Overwrite the snapshot functionality to use the ZISnapshotHelper.
    Forwards all args and kwargs to the QCoDeS Parameter class.

    Args:
        snapshot_cache (ZISnapshotHelper): ZI specific SnapshotHelper object
        zi_node (Node): ZI specific node object of the nodetree
    """

    def __init__(
        self,
        *args,
        snapshot_cache: ZISnapshotHelper,
        zi_node: str,
        tk_node: Node,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.get_raw = kwargs["get_cmd"]
        self.set_raw = kwargs["set_cmd"]
        self.get = self._wrap_get(self.get_raw)
        self.set = self._set_zi
        self._snapshot_cache = snapshot_cache
        self._zi_node = zi_node
        self._tk_node = tk_node

    def __call__(self, *args, **kwargs):
        """Call operator that either gets (empty) or gets the value of a node.

        Args:
            value: Optional value that should be set to the node. If not
                specified the operator will return the value of the node
                instead.
            deep: Flag if the operation should block until the device has
                acknowledged the operation. The operation returns the value
                acknowledged by the device. This takes significantly longer
                than a normal operation and should be used carefully.
            enum: Flag if enumerated values should return the enum value as
                string or return the raw number.
            parse: Flag if the GetParser or SetParser, if present, should be
                applied or not.

        Returns:
            Value of the node for a get operation. If the deep flag is set the
            acknowledged value from the device is returned (applies also for
            the set operation).

        Raises:
            AttributeError: If the connection does not support the necessary
                function to get/set the value.
            RuntimeError: If self.node_info.type if one of the following:
                [ZIPWAWave, ZITriggerSample, ZICntSample, ZIImpedanceSample,
                ZIScopeWave, ZIAuxInSample]. The reason is that these nodes can
                only be polled.
            TypeError: if the deep command is not available for this node
                (e.g. sample nodes)
        """
        if len(args) == 0:
            if self.gettable:
                return self.get(**kwargs)
            raise NotImplementedError("no get cmd found in" + f" Parameter {self.name}")
        if self.settable:
            return self.set(*args, **kwargs)
        raise NotImplementedError("no set cmd found in" + f" Parameter {self.name}")

    def _set_zi(self, *args, **kwargs):
        """ZI specific set that supports returning values.

        QCoDeS does not provide a way to return a value for the set command.
        However a `deep` command in LabOne does return the acknowledged value
        by the device. This function in combination with the custom `__call__`
        method bypasses this problem.

        This function simply wraps around the QCoDeS specific set functionality
        (_wrap_set) stores the returned value by zhinst-toolkit and calls the
        get functionality with the returned value. Thus is acts as a set and
        get within on single command without overwriting the QCoDeS specific
        implementation.
        """
        set_return = None

        def set_wrapper(*args, **kwargs) -> None:
            nonlocal set_return
            set_return = self.set_raw(*args, **kwargs)

        self._wrap_set(set_wrapper)(*args, **kwargs)
        return self._wrap_get(lambda: set_return)() if set_return is not None else None

    def snapshot_base(
        self, update: bool = True, params_to_skip_update: t.List[str] = None
    ) -> dict:
        """State of the parameter as a JSON-compatible dict.

        (everything that the custom JSON encoder class
        :class:`qcodes.utils.helpers.NumpyJSONEncoder` supports)

        If the parameter has been initiated with ``snapshot_value=False``,
        the snapshot will NOT include the ``value`` and ``raw_value`` of the
        parameter.

        Overwrite base class function to use the snapshot_cache.

        Args:
            update: If True, update the state by calling ``parameter.get()``
                unless ``snapshot_get`` of the parameter is ``False``.
                If ``update`` is ``None``, use the current value from the
                ``cache`` unless the cache is invalid. If ``False``, never call
                ``parameter.get()``.
            params_to_skip_update: No effect but may be passed from base class

        Returns:
            base snapshot
        """
        get = self.__dict__["get"]
        try:
            self.get = lambda: self._snapshot_cache.get(self, get)
            return super().snapshot_base(
                update=update, params_to_skip_update=params_to_skip_update
            )
        finally:
            self.get = get

    def subscribe(self) -> None:
        """Subscribe to nodes. Fetch data with the poll command.

        In order to avoid fetching old data that is still in the buffer execute
        a flush command before subscribing to data streams.
        """
        self._tk_node.subscribe()

    def unsubscribe(self) -> None:
        """Unsubscribe data stream.

        Use this command after recording to avoid buffer overflows that may
        increase the latency of other command.
        """
        self._tk_node.unsubscribe()

    def get_as_event(self) -> None:
        """Trigger an event.

        The node data is returned by a subsequent poll command.
        """
        self._tk_node.get_as_event()

    def wait_for_state_change(
        self,
        value: t.Union[int, str],
        *,
        invert: bool = False,
        timeout: float = 2,
        sleep_time: float = 0.005,
    ) -> None:
        """Waits until the node has the expected state/value.

        WARNING: Only supports integer values as reference.

        Args:
            value: expected value of the node.
            invert: Instead of waiting for the value, the function will wait for
                any value except the passed value instead. (default = False)

                Useful when waiting for value to change from existing one.
            timeout: max wait time. (default = 2)
            sleep_time: sleep interval in seconds. (default = 0.006)
        """
        self._tk_node.wait_for_state_change(
            value, invert=invert, timeout=timeout, sleep_time=sleep_time
        )

    @property
    def node_info(self) -> NodeInfo:
        """Zurich Instrument node representation of the Parameter."""
        return self._tk_node.node_info

    @property
    def zi_node(self) -> Node:
        """Zurich Instrument node representation of the Parameter."""
        return self._zi_node

    @property
    def tk_node(self) -> Node:
        """Toolkit node of the Parameter."""
        return self._tk_node


class ZINode(InstrumentChannel):
    """Zurich Instrument specific QCoDeS InstrumentChannel.

    Overwrite the snapshot functionality to use the ZISnapshotHelper.
    Forwards all args and kwargs to the QCoDeS InstrumentChannel class.

    Args:
        snapshot_cache (ZISnapshotHelper): ZI specific SnapshotHelper object
        zi_node (Node): ZI specific node object of the nodetree
    """

    def __init__(
        self,
        *args,
        snapshot_cache: ZISnapshotHelper,
        zi_node: Node = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._snapshot_cache = snapshot_cache
        self._zi_node = zi_node

    def snapshot(self, update: bool = True) -> dict:
        """Decorate a snapshot dictionary with metadata.

        Override base method to make update default True and use the
        ZISnapshotHelper.

        Args:
            update: Passed to snapshot_base. (default = True)

        Returns:
            dict: Base snapshot.
        """
        with self._snapshot_cache.snapshot(self._zi_node) if update else nullcontext():
            return super().snapshot(update)

    def print_readable_snapshot(self, update: bool = True, max_chars: int = 80) -> None:
        """Prints a readable version of the snapshot.

        The readable snapshot includes the name, value and unit of each
        parameter.
        A convenience function to quickly get an overview of the
        status of an instrument.

        Args:
            update: If ``True``, update the state by querying the
                instrument. If ``False``, just use the latest values in memory.
                This argument gets passed to the snapshot function.
            max_chars: the maximum number of characters per line. The
                readable snapshot will be cropped if this value is exceeded.
                Defaults to 80 to be consistent with default terminal width.
        """
        with self._snapshot_cache.snapshot(self._zi_node) if update else nullcontext():
            return super().print_readable_snapshot(update, max_chars)


class ZIChannelList(ChannelList):
    """Zurich Instrument specific QCoDeS InstrumentChannel.

    Overwrite the snapshot functionality to use the ZISnapshotHelper.
    Forwards all args and kwargs to the QCoDeS InstrumentChannel class.

    Args:
        snapshot_cache (ZISnapshotHelper): ZI specific SnapshotHelper object
        zi_node (Node): ZI specific node object of the nodetree
    """

    def __init__(self, *args, snapshot_cache=None, zi_node=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._snapshot_cache = snapshot_cache
        self._zi_node = zi_node

    def snapshot(self, update: bool = True) -> dict:
        """Decorate a snapshot dictionary with metadata.

        Override base method to make update default True and use the
        ZISnapshotHelper.

        Args:
            update: Passed to snapshot_base. (default = True)

        Returns:
            dict: Base snapshot.
        """
        with self._snapshot_cache.snapshot(self._zi_node) if update else nullcontext():
            return super().snapshot(update)

    def print_readable_snapshot(self, update: bool = True, max_chars: int = 80) -> None:
        """Prints a readable version of the snapshot.

        The readable snapshot includes the name, value and unit of each
        parameter.
        A convenience function to quickly get an overview of the
        status of an instrument.

        Args:
            update: If ``True``, update the state by querying the
                instrument. If ``False``, just use the latest values in memory.
                This argument gets passed to the snapshot function.
            max_chars: the maximum number of characters per line. The
                readable snapshot will be cropped if this value is exceeded.
                Defaults to 80 to be consistent with default terminal width.
        """
        with self._snapshot_cache.snapshot(self._zi_node) if update else nullcontext():
            return super().print_readable_snapshot(update, max_chars)


class ZIInstrument(Instrument):
    """Zurich Instrument specific Qcodes Instrument.

    Overwrite the snapshot functionality to use the ZISnapshotHelper.

    Args:
        name: Name of
        snapshot_cache (ZISnapshotHelper): ZI specific SnapshotHelper object
        zi_node (Node): ZI specific node object of the nodetree
    """

    def __init__(self, name, nodetree: NodeTree, is_module=False):
        super().__init__(name)
        self._snapshot_cache = ZISnapshotHelper(nodetree, is_module=is_module)

    def snapshot(self, update: bool = True) -> dict:
        """Decorate a snapshot dictionary with metadata.

        Override base method to make update default True and use the
        ZISnapshotHelper.

        Args:
            update: Passed to snapshot_base.

        Returns:
            dict: Base snapshot.
        """
        with self._snapshot_cache.snapshot() if update else nullcontext():
            return super().snapshot(update)

    def print_readable_snapshot(self, update: bool = True, max_chars: int = 80) -> None:
        """Prints a readable version of the snapshot.

        The readable snapshot includes the name, value and unit of each
        parameter.
        A convenience function to quickly get an overview of the
        status of an instrument.

        Args:
            update: If ``True``, update the state by querying the
                instrument. If ``False``, just use the latest values in memory.
                This argument gets passed to the snapshot function.
            max_chars: the maximum number of characters per line. The
                readable snapshot will be cropped if this value is exceeded.
                Defaults to 80 to be consistent with default terminal width.
        """
        with self._snapshot_cache.snapshot() if update else nullcontext():
            return super().print_readable_snapshot(update, max_chars)


class NodeDict(Mapping):
    """Mapping of dictionary structure results.

    The mapping allows to access data with both the string and the qcodes
    parameter objects.
    """

    def __init__(self, tk_result: t.Union[t.Dict[str, t.Any], TKNodeDict]):
        if not isinstance(tk_result, TKNodeDict):
            self._result = TKNodeDict(tk_result)
        else:
            self._result = tk_result

    def __repr__(self):
        return repr(self._result)

    def __getitem__(self, key: t.Union[str, ZIParameter]):
        if isinstance(key, ZIParameter):
            return self._result[key.zi_node.lower()]
        return self._result[key]

    def __iter__(self):
        return iter(self._result)

    def __len__(self):
        return len(self._result)

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Convert the WildcardResult to a dictionary.

        After conversion, :class:`Node` objects cannot be used to get items.
        """
        return self._result.to_dict()


def tk_node_to_qcodes_list(tk_node: Node) -> t.List[str]:
    """Convert a toolkit node to a list of elements that form a QCoDeS object.

    Args:
        tk_node: Toolkit node to convert.

    Return:
        List of strings that form a QCoDeS object.
    """
    if tk_node.raw_tree[-1].isdigit():
        parents = tk_node.raw_tree
        name = "value"
    else:
        parents = tk_node.raw_tree[:-1]
        name = tk_node.raw_tree[-1]
        # Attributes are not allowed to start with a number (#31)
        name = "_" + name if name[0].isdigit() else name
    parents = list(parents)
    numbers = [subnode for subnode in parents if subnode.isdigit()]
    while numbers:
        number = numbers.pop()
        index = parents.index(number)
        parents[index - 1] = parents[index - 1] + number
        parents.pop(index)
        if not numbers:
            numbers = [subnode for subnode in parents if subnode.isdigit()]
    parents.append(name)
    return parents


def tk_node_to_parameter(root: t.Any, tk_node: Node) -> t.Any:
    """Convert a Toolkit node into a QCoDeS Parameter.

    Args:
        root: Root from which the node should be derived.
        tk_node: Toolkit node to convert.

    Returns:
        QCoDeS Parameter that matches the given tk node.
    """
    qcodes_list = list(tk_node.raw_tree)
    if qcodes_list[-1].isdigit():
        qcodes_list.append("value")
    current_layer = root
    for element in qcodes_list[:-1]:
        qcodes_list = tk_node_to_qcodes_list(tk_node)
        if element.isdigit():
            current_layer = current_layer[int(element)]
        else:
            current_layer = current_layer.submodules[element]
    return current_layer.parameters[qcodes_list[-1]]


def _get_submodule(
    layer, parents: t.List[str], snapshot_cache: ZISnapshotHelper
) -> ZINode:
    """Get the nested parent element for a node.

    Reuse existing subnodes and automatically create them if they don`t
    exist.

    Args:
        parents: Nested parents of a node as str.
        snapshot_cache: Object of the snapshot cache.

    Returns:
        ZINode: direct parent of the node
    """
    weird_nodes = ["tamp0", "tamp1"]
    current_layer = layer
    for i, node in enumerate(parents):
        if node[-1].isdigit() and node not in weird_nodes:
            offset = 0
            for char in reversed(node):
                if char.isdigit():
                    offset += 1
                else:
                    break
            number = int(node[-offset:])
            name = node[:-offset]
            if not current_layer.submodules or name not in current_layer.submodules:
                # create channel_list
                channel_list = ZIChannelList(
                    current_layer,
                    name,
                    ZINode,
                    zi_node="/".join(parents[:i] + [name]),
                    snapshot_cache=snapshot_cache,
                )
                current_layer.add_submodule(name, channel_list)
            if len(current_layer.submodules[name]) <= number:
                # Add new items to list until the required length is reached. (#31)
                current_length = len(current_layer.submodules[name])
                for item in range(number - current_length + 1):
                    module = ZINode(
                        current_layer,
                        name + str(current_length + item),
                        zi_node="/".join(
                            parents[:i] + [name, str(current_length + item)]
                        ),
                        snapshot_cache=snapshot_cache,
                    )
                    current_layer.submodules[name].append(module)
            current_layer = current_layer.submodules[name][number]
        elif node not in current_layer.submodules:
            module = ZINode(
                current_layer,
                node,
                zi_node="/".join(parents[: i + 1]),
                snapshot_cache=snapshot_cache,
            )
            current_layer.add_submodule(node, module)
            current_layer = module
        else:
            current_layer = current_layer.submodules.get(node)
    return current_layer


def init_nodetree(
    layer,
    nodetree: NodeTree,
    snapshot_cache: ZISnapshotHelper,
    blacklist: tuple = tuple(),
) -> None:
    """Generate nested qcodes parameter from the device nodetree.

    Args:
        layer: current layer in the nodetree.
        nodetree: underlying toolkit node tree.
        snapshot_cache: Instance of the SnapshotHelper.
        blacklist: nodes to be blacklisted.
    """
    snapshot_blacklist = ["fwlog", "values"]

    is_complex = re.compile("demods/./sample")

    for node, info in nodetree:
        if info.get("Node", "") in blacklist:
            continue
        try:
            qcodes_list = tk_node_to_qcodes_list(node)
            name = qcodes_list[-1]
            parent = _get_submodule(layer, qcodes_list[:-1], snapshot_cache)
            do_snapshot = (
                "Stream" not in info.get("Properties")
                and "ZIVector" not in info.get("Type")
                and "Read" in info.get("Properties")
                and not any(x in node.raw_tree for x in snapshot_blacklist)
            )
            name = name + "_" if hasattr(parent, name) else name
            parent.add_parameter(
                parameter_class=ZIParameter,
                name=name,
                docstring=info.get("Description"),
                unit=(
                    info.get("Unit")
                    if info.get("Unit") not in ["None", "Dependent"]
                    else None
                ),
                get_cmd=node._get,
                set_cmd=node._set,
                vals=(
                    ComplexNumbers()
                    if re.match(is_complex, info.get("Node").lower())
                    else None
                ),
                snapshot_value=do_snapshot,
                snapshot_get=do_snapshot,
                zi_node=info.get("Node"),
                tk_node=node,
                snapshot_cache=snapshot_cache,
            )
        except ValueError as e:
            print(f"Node {info.get('Node')} could not be added as parameter\n", e)



import yaml
import os
from glob import glob
from pathlib import Path

EXAMPLES_DIRECTORY = "examples/"


def test_config_existence():
    assert os.path.exists(EXAMPLES_DIRECTORY + "test.spec.yml")


def test_example_config():
    with open(EXAMPLES_DIRECTORY + "test.spec.yml", "rb") as f:
        test_spec = yaml.load(f, Loader=yaml.Loader)

    examples_list = glob(EXAMPLES_DIRECTORY + "*.md")

    for example in examples_list:
        example_name = Path(example).name.split(".")[0]
        if example_name == "README":
            continue

        # Check that the notebook is in the config file
        assert (
            example_name in test_spec
        ), f'Example "{example_name}" was not included in the configuration file.'

        # Check that it has the device id specified
        assert (
            "device_type" in test_spec[example_name]
        ), f'No device_type was specified for the example "{example_name}".'

        # Check that the device id was written
        assert (
            test_spec[example_name]["device_type"] != None
        ), f'The field "device_type" is empty for the example "{example_name}".'



"""Device instance module.

The classes can be used without creating as session to a data server.
"""
import typing as t

from zhinst.qcodes.driver.devices.base import ZIBaseInstrument
from zhinst.qcodes.driver.devices.hdawg import HDAWG as HDAWGDriver
from zhinst.qcodes.driver.devices.pqsc import PQSC as PQSCDriver
from zhinst.qcodes.driver.devices.shfqa import SHFQA as SHFQADriver
from zhinst.qcodes.driver.devices.shfqc import SHFQC as SHFQCDriver
from zhinst.qcodes.driver.devices.shfsg import SHFSG as SHFSGDriver
from zhinst.qcodes.driver.devices.uhfli import UHFLI as UHFLIDriver
from zhinst.qcodes.driver.devices.uhfqa import UHFQA as UHFQADriver
from zhinst.qcodes.session import ZISession


class ZIDevice(ZIBaseInstrument):
    """QCoDeS driver for the Zurich Instruments ZIDevice.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class SHFQA(SHFQADriver):
    """QCoDeS driver for the Zurich Instruments SHFQA.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class SHFSG(SHFSGDriver):
    """QCoDeS driver for the Zurich Instruments SHFSG.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class HDAWG(HDAWGDriver):
    """QCoDeS driver for the Zurich Instruments HDAWG.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class PQSC(PQSCDriver):
    """QCoDeS driver for the Zurich Instruments PQSC.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class SHFQC(SHFQCDriver):
    """QCoDeS driver for the Zurich Instruments SHFQC.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class UHFLI(UHFLIDriver):
    """QCoDeS driver for the Zurich Instruments UHFLI.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class UHFQA(UHFQADriver):
    """QCoDeS driver for the Zurich Instruments UHFQA.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class MFLI(ZIBaseInstrument):
    """QCoDeS driver for the Zurich Instruments MFLI.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class MFIA(ZIBaseInstrument):
    """QCoDeS driver for the Zurich Instruments MFIA.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8004)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8004,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=False, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self


class HF2(ZIBaseInstrument):
    """QCoDeS driver for the Zurich Instruments HF2.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port. (default = 8005)
        interface: Device interface (e.g. = "1GbE"). If not specified
            the default interface from the discover is used.
        name: Name of the instrument in qcodes.
        raw: Flag if qcodes instance should only created with the nodes and
            not forwarding the toolkit functions. (default = False)
        new_session: By default zhinst-qcodes reuses already existing data
            server session (within itself only), meaning only one session to a
            data server exists. Setting the flag will create a new session.

    Warning:
        Creating a new session should be done carefully and reusing
        the created session is not possible. Consider instantiating a
        new session directly.
    """

    def __init__(
        self,
        serial: str,
        host: str,
        port: int = 8005,
        *,
        interface: t.Optional[str] = None,
        name=None,
        raw=False,
        new_session: bool = False,
    ):
        session = ZISession(host, port, hf2=True, new_session=new_session)
        tk_device = session.toolkit_session.connect_device(serial, interface=interface)
        super().__init__(tk_device, session, name=name, raw=raw)
        session.devices[self.serial] = self
